@@ -34,8 +34,20 @@ extern "C" {
 #include <stdlib.h>
 #include <stdint.h>
 #include <string.h>
+#include <math.h>
+#include <float.h>
 
 #include <spa/utils/defs.h>
+#include <spa/utils/string.h>
+
+/** \defgroup spa_json JSON
+ * Relaxed JSON variant parsing
+ */
+
+/**
+ * \addtogroup spa_json
+ * \{
+ */
 
 /* a simple JSON compatible tokenizer */
 struct spa_json {
@@ -59,6 +71,8 @@ static inline void spa_json_enter(struct spa_json * iter, struct spa_json * sub)
 	*sub = SPA_JSON_ENTER(iter);
 }
 
+#define SPA_JSON_SAVE(iter) (struct spa_json) { (iter)->cur, (iter)->end, }
+
 /** Get the next token. \a value points to the token and the return value
  * is the length. */
 static inline int spa_json_next(struct spa_json * iter, const char **value)
@@ -77,7 +91,7 @@ static inline int spa_json_next(struct spa_json * iter, const char **value)
 			goto again;
 		case __STRUCT:
 			switch (cur) {
-			case '\t': case ' ': case '\r': case '\n': case ':': case '=': case ',':
+			case '\0': case '\t': case ' ': case '\r': case '\n': case ':': case '=': case ',':
 				continue;
 			case '#':
 				iter->state = __COMMENT;
@@ -124,8 +138,7 @@ static inline int spa_json_next(struct spa_json * iter, const char **value)
 				iter->state = __STRUCT;
 				if (iter->depth > 0)
 					continue;
-				iter->cur++;
-				return iter->cur - *value;
+				return ++iter->cur - *value;
 			case 240 ... 247:
 				utf8_remain++;
 				SPA_FALLTHROUGH;
@@ -227,9 +240,10 @@ static inline bool spa_json_is_null(const char *val, int len)
 static inline int spa_json_parse_float(const char *val, int len, float *result)
 {
 	char *end;
-	*result = strtof(val, &end);
-	return end == val + len;
+	*result = spa_strtof(val, &end);
+	return len > 0 && end == val + len;
 }
+
 static inline bool spa_json_is_float(const char *val, int len)
 {
 	float dummy;
@@ -244,6 +258,40 @@ static inline int spa_json_get_float(struct spa_json *iter, float *res)
 	return spa_json_parse_float(value, len, res);
 }
 
+static inline char *spa_json_format_float(char *str, int size, float val)
+{
+	if (SPA_UNLIKELY(!isnormal(val))) {
+		if (val == INFINITY)
+			val = FLT_MAX;
+		else if (val == -INFINITY)
+			val = FLT_MIN;
+		else
+			val = 0.0f;
+	}
+	return spa_dtoa(str, size, val);
+}
+
+/* int */
+static inline int spa_json_parse_int(const char *val, int len, int *result)
+{
+	char *end;
+	*result = strtol(val, &end, 0);
+	return len > 0 && end == val + len;
+}
+static inline bool spa_json_is_int(const char *val, int len)
+{
+	int dummy;
+	return spa_json_parse_int(val, len, &dummy);
+}
+static inline int spa_json_get_int(struct spa_json *iter, int *res)
+{
+	const char *value;
+	int len;
+	if ((len = spa_json_next(iter, &value)) <= 0)
+		return -1;
+	return spa_json_parse_int(value, len, res);
+}
+
 /* bool */
 static inline bool spa_json_is_true(const char *val, int len)
 {
@@ -283,11 +331,14 @@ static inline bool spa_json_is_string(const char *val, int len)
 	return len > 1 && *val == '"';
 }
 
-static inline int spa_json_parse_string(const char *val, int len, char *result)
+static inline int spa_json_parse_stringn(const char *val, int len, char *result, int maxlen)
 {
 	const char *p;
+	if (maxlen <= len)
+		return -1;
 	if (!spa_json_is_string(val, len)) {
-		strncpy(result, val, len);
+		if (result != val)
+			strncpy(result, val, len);
 		result += len;
 	} else {
 		for (p = val+1; p < val + len; p++) {
@@ -303,7 +354,18 @@ static inline int spa_json_parse_string(const char *val, int len, char *result)
 					*result++ = '\t';
 				else if (*p == 'f')
 					*result++ = '\f';
-				else
+				else if (*p == 'u') {
+					char *end;
+					uint16_t v = strtol(p+1, &end, 16);
+					if (p+1 == end) {
+						*result++ = *p;
+					} else {
+						p = end-1;
+						if (v > 0xff)
+							*result++ = (v >> 8) & 0xff;
+						*result++ = v & 0xff;
+					}
+				} else
 					*result++ = *p;
 			} else if (*p == '\"') {
 				break;
@@ -315,18 +377,24 @@ static inline int spa_json_parse_string(const char *val, int len, char *result)
 	return 1;
 }
 
+static inline int spa_json_parse_string(const char *val, int len, char *result)
+{
+	return spa_json_parse_stringn(val, len, result, len+1);
+}
+
 static inline int spa_json_get_string(struct spa_json *iter, char *res, int maxlen)
 {
 	const char *value;
 	int len;
-	if ((len = spa_json_next(iter, &value)) <= 0 || maxlen < len)
+	if ((len = spa_json_next(iter, &value)) <= 0)
 		return -1;
-	return spa_json_parse_string(value, len, res);
+	return spa_json_parse_stringn(value, len, res, maxlen);
 }
 
 static inline int spa_json_encode_string(char *str, int size, const char *val)
 {
 	int len = 0;
+	static const char hex[] = { "0123456789abcdef" };
 #define __PUT(c) { if (len < size) *str++ = c; len++; }
 	__PUT('"');
 	while (*val) {
@@ -346,8 +414,18 @@ static inline int spa_json_encode_string(char *str, int size, const char *val)
 		case '\f':
 			__PUT('\\'); __PUT('f');
 			break;
+		case '\\':
+		case '"':
+			__PUT('\\'); __PUT(*val);
+			break;
 		default:
-			__PUT(*val);
+			if (*val > 0 && *val < 0x20) {
+				__PUT('\\'); __PUT('u');
+				__PUT('0'); __PUT('0');
+				__PUT(hex[((*val)>>4)&0xf]); __PUT(hex[(*val)&0xf]);
+			} else {
+				__PUT(*val);
+			}
 			break;
 		}
 		val++;
@@ -358,6 +436,10 @@ static inline int spa_json_encode_string(char *str, int size, const char *val)
 	return len-1;
 }
 
+/**
+ * \}
+ */
+
 #ifdef __cplusplus
 } /* extern "C" */
 #endif
