@@ -72,11 +72,9 @@ error:
 	return NULL;
 }
 
-static void release_card(uint32_t index)
+static void release_card(struct card *c)
 {
-	struct card *c;
-	if ((c = find_card(index)) == NULL)
-		return;
+	spa_assert(c->ref > 0);
 
 	if (--c->ref > 0)
 		return;
@@ -104,11 +102,17 @@ static int alsa_set_param(struct state *state, const char *k, const char *s)
 	} else if (spa_streq(k, SPA_KEY_AUDIO_POSITION)) {
 		spa_alsa_parse_position(&state->default_pos, s, strlen(s));
 		fmt_change++;
+	} else if (spa_streq(k, SPA_KEY_AUDIO_ALLOWED_RATES)) {
+		state->n_allowed_rates = spa_alsa_parse_rates(state->allowed_rates,
+				MAX_RATES, s, strlen(s));
+		fmt_change++;
 	} else if (spa_streq(k, "iec958.codecs")) {
 		spa_alsa_parse_iec958_codecs(&state->iec958_codecs, s, strlen(s));
 		fmt_change++;
 	} else if (spa_streq(k, "api.alsa.period-size")) {
 		state->default_period_size = atoi(s);
+	} else if (spa_streq(k, "api.alsa.period-num")) {
+		state->default_period_num = atoi(s);
 	} else if (spa_streq(k, "api.alsa.headroom")) {
 		state->default_headroom = atoi(s);
 	} else if (spa_streq(k, "api.alsa.start-delay")) {
@@ -121,6 +125,10 @@ static int alsa_set_param(struct state *state, const char *k, const char *s)
 		state->props.use_chmap = spa_atob(s);
 	} else if (spa_streq(k, "api.alsa.multi-rate")) {
 		state->multi_rate = spa_atob(s);
+	} else if (spa_streq(k, "latency.internal.rate")) {
+		state->process_latency.rate = atoi(s);
+	} else if (spa_streq(k, "latency.internal.ns")) {
+		state->process_latency.ns = atoi(s);
 	} else if (spa_streq(k, "clock.name")) {
 		spa_scnprintf(state->clock_name,
 				sizeof(state->clock_name), "%s", s);
@@ -152,6 +160,22 @@ static int position_to_string(struct channel_map *map, char *val, size_t len)
 	return 0;
 }
 
+static int uint32_array_to_string(uint32_t *vals, uint32_t n_vals, char *val, size_t len)
+{
+	uint32_t i, o = 0;
+	int r;
+	o += snprintf(val, len, "[ ");
+	for (i = 0; i < n_vals; i++) {
+		r = snprintf(val+o, len-o, "%s%d", i == 0 ? "" : ", ", vals[i]);
+		if (r < 0 || o + r >= len)
+			return -ENOSPC;
+		o += r;
+	}
+	if (len > o)
+		o += snprintf(val+o, len-o, " ]");
+	return 0;
+}
+
 struct spa_pod *spa_alsa_enum_propinfo(struct state *state,
 		uint32_t idx, struct spa_pod_builder *b)
 {
@@ -197,62 +221,100 @@ struct spa_pod *spa_alsa_enum_propinfo(struct state *state,
 		break;
 	}
 	case 4:
+	{
+		char buf[1024];
+		uint32_array_to_string(state->allowed_rates, state->n_allowed_rates, buf, sizeof(buf));
+		param = spa_pod_builder_add_object(b,
+			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
+			SPA_PROP_INFO_name, SPA_POD_String(SPA_KEY_AUDIO_ALLOWED_RATES),
+			SPA_PROP_INFO_description, SPA_POD_String("Audio Allowed Rates"),
+			SPA_PROP_INFO_type, SPA_POD_String(buf),
+			SPA_PROP_INFO_params, SPA_POD_Bool(true));
+		break;
+	}
+	case 5:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
 			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.period-size"),
 			SPA_PROP_INFO_description, SPA_POD_String("Period Size"),
-			SPA_PROP_INFO_type, SPA_POD_Int(state->default_period_size),
-			SPA_PROP_INFO_params, SPA_POD_Bool(true));
-		break;
-	case 5:
-		param = spa_pod_builder_add_object(b,
-			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
-			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.headroom"),
-			SPA_PROP_INFO_description, SPA_POD_String("Headroom"),
-			SPA_PROP_INFO_type, SPA_POD_Int(state->default_headroom),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Int(state->default_period_size, 0, 8192),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
 	case 6:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
-			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.start-delay"),
-			SPA_PROP_INFO_description, SPA_POD_String("Start Delay"),
-			SPA_PROP_INFO_type, SPA_POD_Int(state->default_start_delay),
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.period-num"),
+			SPA_PROP_INFO_description, SPA_POD_String("Number of Periods"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Int(state->default_period_num, 0, 1024),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
 	case 7:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
-			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.disable-mmap"),
-			SPA_PROP_INFO_description, SPA_POD_String("Disable MMAP"),
-			SPA_PROP_INFO_type, SPA_POD_Bool(state->disable_mmap),
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.headroom"),
+			SPA_PROP_INFO_description, SPA_POD_String("Headroom"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Int(state->default_headroom, 0, 8192),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
 	case 8:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
-			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.disable-batch"),
-			SPA_PROP_INFO_description, SPA_POD_String("Disable Batch"),
-			SPA_PROP_INFO_type, SPA_POD_Bool(state->disable_batch),
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.start-delay"),
+			SPA_PROP_INFO_description, SPA_POD_String("Start Delay"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Int(state->default_start_delay, 0, 8192),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
 	case 9:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
-			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.use-chmap"),
-			SPA_PROP_INFO_description, SPA_POD_String("Use the driver channelmap"),
-			SPA_PROP_INFO_type, SPA_POD_Bool(state->props.use_chmap),
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.disable-mmap"),
+			SPA_PROP_INFO_description, SPA_POD_String("Disable MMAP"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(state->disable_mmap),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
 	case 10:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
-			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.multi-rate"),
-			SPA_PROP_INFO_description, SPA_POD_String("Support multiple rates"),
-			SPA_PROP_INFO_type, SPA_POD_Bool(state->multi_rate),
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.disable-batch"),
+			SPA_PROP_INFO_description, SPA_POD_String("Disable Batch"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(state->disable_batch),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
 	case 11:
+		param = spa_pod_builder_add_object(b,
+			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.use-chmap"),
+			SPA_PROP_INFO_description, SPA_POD_String("Use the driver channelmap"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(state->props.use_chmap),
+			SPA_PROP_INFO_params, SPA_POD_Bool(true));
+		break;
+	case 12:
+		param = spa_pod_builder_add_object(b,
+			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.multi-rate"),
+			SPA_PROP_INFO_description, SPA_POD_String("Support multiple rates"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(state->multi_rate),
+			SPA_PROP_INFO_params, SPA_POD_Bool(true));
+		break;
+	case 13:
+		param = spa_pod_builder_add_object(b,
+			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
+			SPA_PROP_INFO_name, SPA_POD_String("latency.internal.rate"),
+			SPA_PROP_INFO_description, SPA_POD_String("Internal latency in samples"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Int(state->process_latency.rate,
+				0, 65536),
+			SPA_PROP_INFO_params, SPA_POD_Bool(true));
+		break;
+	case 14:
+		param = spa_pod_builder_add_object(b,
+			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
+			SPA_PROP_INFO_name, SPA_POD_String("latency.internal.ns"),
+			SPA_PROP_INFO_description, SPA_POD_String("Internal latency in nanoseconds"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Long(state->process_latency.ns,
+				0, 2 * SPA_NSEC_PER_SEC),
+			SPA_PROP_INFO_params, SPA_POD_Bool(true));
+		break;
+	case 15:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
 			SPA_PROP_INFO_name, SPA_POD_String("clock.name"),
@@ -289,9 +351,17 @@ int spa_alsa_add_prop_params(struct state *state, struct spa_pod_builder *b)
 	spa_pod_builder_string(b, SPA_KEY_AUDIO_POSITION);
 	spa_pod_builder_string(b, buf);
 
+	uint32_array_to_string(state->allowed_rates, state->n_allowed_rates,
+			buf, sizeof(buf));
+	spa_pod_builder_string(b, SPA_KEY_AUDIO_ALLOWED_RATES);
+	spa_pod_builder_string(b, buf);
+
 	spa_pod_builder_string(b, "api.alsa.period-size");
 	spa_pod_builder_int(b, state->default_period_size);
 
+	spa_pod_builder_string(b, "api.alsa.period-num");
+	spa_pod_builder_int(b, state->default_period_num);
+
 	spa_pod_builder_string(b, "api.alsa.headroom");
 	spa_pod_builder_int(b, state->default_headroom);
 
@@ -310,6 +380,12 @@ int spa_alsa_add_prop_params(struct state *state, struct spa_pod_builder *b)
 	spa_pod_builder_string(b, "api.alsa.multi-rate");
 	spa_pod_builder_bool(b, state->multi_rate);
 
+	spa_pod_builder_string(b, "latency.internal.rate");
+	spa_pod_builder_int(b, state->process_latency.rate);
+
+	spa_pod_builder_string(b, "latency.internal.ns");
+	spa_pod_builder_long(b, state->process_latency.ns);
+
 	spa_pod_builder_string(b, "clock.name");
 	spa_pod_builder_string(b, state->clock_name);
 
@@ -343,8 +419,11 @@ int spa_alsa_parse_prop_params(struct state *state, struct spa_pod *params)
 		if (spa_pod_is_string(pod)) {
 			spa_pod_copy_string(pod, sizeof(value), value);
 		} else if (spa_pod_is_int(pod)) {
-			snprintf(value, sizeof(value), "%u",
+			snprintf(value, sizeof(value), "%d",
 					SPA_POD_VALUE(struct spa_pod_int, pod));
+		} else if (spa_pod_is_long(pod)) {
+			snprintf(value, sizeof(value), "%"PRIi64,
+					SPA_POD_VALUE(struct spa_pod_long, pod));
 		} else if (spa_pod_is_bool(pod)) {
 			snprintf(value, sizeof(value), "%s",
 					SPA_POD_VALUE(struct spa_pod_bool, pod) ?
@@ -352,7 +431,7 @@ int spa_alsa_parse_prop_params(struct state *state, struct spa_pod *params)
 		} else
 			continue;
 
-		spa_log_debug(state->log, "key:'%s' val:'%s'", name, value);
+		spa_log_info(state->log, "key:'%s' val:'%s'", name, value);
 		alsa_set_param(state, name, value);
 		changed++;
 	}
@@ -369,6 +448,7 @@ int spa_alsa_init(struct state *state, const struct spa_dict *info)
 
 	snd_config_update_free_global();
 
+	state->multi_rate = true;
 	for (i = 0; info && i < info->n_items; i++) {
 		const char *k = info->items[i].key;
 		const char *s = info->items[i].value;
@@ -378,17 +458,15 @@ int spa_alsa_init(struct state *state, const struct spa_dict *info)
 			state->card_index = atoi(s);
 		} else if (spa_streq(k, SPA_KEY_API_ALSA_OPEN_UCM)) {
 			state->open_ucm = spa_atob(s);
-		} else if (spa_streq(k, "latency.internal.rate")) {
-			state->process_latency.rate = atoi(s);
-		} else if (spa_streq(k, "latency.internal.ns")) {
-			state->process_latency.ns = atoi(s);
+		} else if (spa_streq(k, "clock.quantum-limit")) {
+			spa_atou32(s, &state->quantum_limit, 0);
 		} else {
 			alsa_set_param(state, k, s);
 		}
 	}
 	if (state->clock_name[0] == '\0')
 		snprintf(state->clock_name, sizeof(state->clock_name),
-				"api.alsa.%d", state->card_index);
+				"api.alsa.%u", state->card_index);
 
 	if (state->stream == SND_PCM_STREAM_PLAYBACK) {
 		state->is_iec958 = spa_strstartswith(state->props.device, "iec958");
@@ -398,7 +476,7 @@ int spa_alsa_init(struct state *state, const struct spa_dict *info)
 
 	state->card = ensure_card(state->card_index, state->open_ucm);
 	if (state->card == NULL) {
-		spa_log_error(state->log, "can't create card %d", state->card_index);
+		spa_log_error(state->log, "can't create card %u", state->card_index);
 		return -errno;
 	}
 	return 0;
@@ -406,8 +484,11 @@ int spa_alsa_init(struct state *state, const struct spa_dict *info)
 
 int spa_alsa_clear(struct state *state)
 {
-	release_card(state->card_index);
+	release_card(state->card);
+
 	state->card = NULL;
+	state->card_index = SPA_ID_INVALID;
+
 	return 0;
 }
 
@@ -659,6 +740,15 @@ static void sanitize_map(snd_pcm_chmap_t* map)
 	}
 }
 
+static bool uint32_array_contains(uint32_t *vals, uint32_t n_vals, uint32_t val)
+{
+	uint32_t i;
+	for (i = 0; i < n_vals; i++)
+		if (vals[i] == val)
+			return true;
+	return false;
+}
+
 static int add_rate(struct state *state, uint32_t scale, bool all, uint32_t index, uint32_t *next,
 		snd_pcm_hw_params_t *params, struct spa_pod_builder *b)
 {
@@ -671,29 +761,55 @@ static int add_rate(struct state *state, uint32_t scale, bool all, uint32_t inde
 	CHECK(snd_pcm_hw_params_get_rate_min(params, &min, &dir), "get_rate_min");
 	CHECK(snd_pcm_hw_params_get_rate_max(params, &max, &dir), "get_rate_max");
 
-	rate = state->default_rate;
 	if (!state->multi_rate && state->card->format_ref > 0)
 		rate = state->card->rate;
+	else
+		rate = state->default_rate;
 
-	if (rate != 0 && !all) {
-		if (min < rate)
-			min = rate;
-		if (max > rate)
-			max = rate;
-	}
+	if (rate < min || rate > max)
+		rate = 0;
+
+	if (rate != 0 && !all)
+		min = max = rate;
+
+	if (rate == 0)
+		rate = state->position ? state->position->clock.rate.denom : DEFAULT_RATE;
+
+	rate = SPA_CLAMP(rate, min, max);
 
 	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_rate, 0);
 
 	spa_pod_builder_push_choice(b, &f[0], SPA_CHOICE_None, 0);
 	choice = (struct spa_pod_choice*)spa_pod_builder_frame(b, &f[0]);
 
-	rate = state->position ? state->position->clock.rate.denom : DEFAULT_RATE;
+	if (state->n_allowed_rates > 0) {
+		uint32_t i, v, last = 0, count = 0;
 
-	spa_pod_builder_int(b, SPA_CLAMP(rate, min, max) * scale);
-	if (min != max) {
-		spa_pod_builder_int(b, min * scale);
-		spa_pod_builder_int(b, max * scale);
-		choice->body.type = SPA_CHOICE_Range;
+		if (uint32_array_contains(state->allowed_rates, state->n_allowed_rates, rate)) {
+			spa_pod_builder_int(b, rate * scale);
+			count++;
+		}
+		for (i = 0; i < state->n_allowed_rates; i++) {
+			v = SPA_CLAMP(state->allowed_rates[i], min, max);
+			if (v != last &&
+			    uint32_array_contains(state->allowed_rates, state->n_allowed_rates, v)) {
+				spa_pod_builder_int(b, v * scale);
+				if (count == 0)
+					spa_pod_builder_int(b, v * scale);
+				count++;
+			}
+			last = v;
+		}
+		if (count > 1)
+			choice->body.type = SPA_CHOICE_Enum;
+	} else {
+		spa_pod_builder_int(b, rate * scale);
+
+		if (min != max) {
+			spa_pod_builder_int(b, min * scale);
+			spa_pod_builder_int(b, max * scale);
+			choice->body.type = SPA_CHOICE_Range;
+		}
 	}
 	spa_pod_builder_pop(b, &f[0]);
 
@@ -1162,7 +1278,11 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 		case SPA_AUDIO_IEC958_CODEC_MPEG2_AAC:
 			break;
 		case SPA_AUDIO_IEC958_CODEC_EAC3:
-			rrate *= 4;
+			/* EAC3 has 3 rates, 32, 44.1 and 48KHz. We need to
+			 * open the device in 4x that rate. Some clients
+			 * already multiply (mpv,..) others don't (vlc). */
+			if (rrate <= 48000)
+				rrate *= 4;
 			break;
 		case SPA_AUDIO_IEC958_CODEC_TRUEHD:
 		case SPA_AUDIO_IEC958_CODEC_DTSHD:
@@ -1322,25 +1442,43 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 		state->card->rate = rrate;
 
 	dir = 0;
-	period_size = state->default_period_size ? state->default_period_size : 1024;
+	period_size = state->default_period_size;
 	is_batch = snd_pcm_hw_params_is_batch(params) &&
 		!state->disable_batch;
 
 	if (is_batch) {
+		if (period_size == 0)
+			period_size = state->position ? state->position->clock.duration : DEFAULT_PERIOD;
+		if (period_size == 0)
+			period_size = DEFAULT_PERIOD;
 		/* batch devices get their hw pointers updated every period. Make
-		 * the period smaller and add one period of headroom */
-		period_size /= 2;
+		 * the period smaller and add one period of headroom. Limit the
+		 * period size to our default so that we don't create too much
+		 * headroom. */
+		period_size = SPA_MIN(period_size, DEFAULT_PERIOD) / 2;
 		spa_log_info(state->log, "%s: batch mode, period_size:%ld",
 			state->props.device, period_size);
 	} else {
+		if (period_size == 0)
+			period_size = DEFAULT_PERIOD;
 		/* disable ALSA wakeups, we use a timer */
 		if (snd_pcm_hw_params_can_disable_period_wakeup(params))
 			CHECK(snd_pcm_hw_params_set_period_wakeup(hndl, params, 0), "set_period_wakeup");
 	}
 
 	CHECK(snd_pcm_hw_params_set_period_size_near(hndl, params, &period_size, &dir), "set_period_size_near");
-	CHECK(snd_pcm_hw_params_get_buffer_size_max(params, &state->buffer_frames), "get_buffer_size_max");
-	CHECK(snd_pcm_hw_params_set_buffer_size_near(hndl, params, &state->buffer_frames), "set_buffer_size_near");
+
+	state->period_frames = period_size;
+
+	if (state->default_period_num != 0) {
+		periods = state->default_period_num;
+		CHECK(snd_pcm_hw_params_set_periods_near(hndl, params, &periods, &dir), "set_periods");
+		state->buffer_frames = period_size * periods;
+	} else {
+		CHECK(snd_pcm_hw_params_get_buffer_size_max(params, &state->buffer_frames), "get_buffer_size_max");
+		CHECK(snd_pcm_hw_params_set_buffer_size_near(hndl, params, &state->buffer_frames), "set_buffer_size_near");
+		periods = state->buffer_frames / period_size;
+	}
 
 	state->headroom = state->default_headroom;
 	if (is_batch)
@@ -1352,9 +1490,6 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	state->latency[state->port_direction].min_rate = state->headroom;
 	state->latency[state->port_direction].max_rate = state->headroom;
 
-	state->period_frames = period_size;
-	periods = state->buffer_frames / state->period_frames;
-
 	spa_log_info(state->log, "%s (%s): format:%s access:%s-%s rate:%d channels:%d "
 			"buffer frames %lu, period frames %lu, periods %u, frame_size %zd "
 			"headroom %u start-delay:%u",
@@ -1383,8 +1518,10 @@ static int set_swparams(struct state *state)
 	/* get the current params */
 	CHECK(snd_pcm_sw_params_current(hndl, params), "sw_params_current");
 
-	CHECK(snd_pcm_sw_params_set_tstamp_mode(hndl, params, SND_PCM_TSTAMP_ENABLE), "sw_params_set_tstamp_mode");
-
+	CHECK(snd_pcm_sw_params_set_tstamp_mode(hndl, params, SND_PCM_TSTAMP_ENABLE),
+			"sw_params_set_tstamp_mode");
+	CHECK(snd_pcm_sw_params_set_tstamp_type(hndl, params, SND_PCM_TSTAMP_TYPE_MONOTONIC),
+			"sw_params_set_tstamp_type");
 #if 0
 	snd_pcm_uframes_t boundary;
 	CHECK(snd_pcm_sw_params_get_boundary(params, &boundary), "get_boundary");
@@ -1511,6 +1648,9 @@ static int alsa_recover(struct state *state, int err)
 	case SND_PCM_STATE_SUSPENDED:
 		spa_log_info(state->log, "%s: recover from state %s",
 				state->props.device, snd_pcm_state_name(st));
+		res = snd_pcm_resume(state->hndl);
+		if (res >= 0)
+		        return res;
 		err = -ESTRPIPE;
 		break;
 	default:
@@ -1535,10 +1675,10 @@ recover:
 	return do_start(state);
 }
 
-static int get_status(struct state *state, snd_pcm_uframes_t *delay, snd_pcm_uframes_t *target)
+static int get_avail(struct state *state, uint64_t current_time)
 {
-	snd_pcm_sframes_t avail;
 	int res;
+	snd_pcm_sframes_t avail;
 
 	if (SPA_UNLIKELY((avail = snd_pcm_avail(state->hndl)) < 0)) {
 		if ((res = alsa_recover(state, avail)) < 0)
@@ -1551,6 +1691,48 @@ static int get_status(struct state *state, snd_pcm_uframes_t *delay, snd_pcm_ufr
 	} else {
 		state->alsa_recovering = false;
 	}
+	return avail;
+}
+
+#if 0
+static int get_avail_htimestamp(struct state *state, uint64_t current_time)
+{
+	int res;
+	snd_pcm_uframes_t avail;
+	snd_htimestamp_t tstamp;
+	uint64_t then;
+
+	if ((res = snd_pcm_htimestamp(state->hndl, &avail, &tstamp)) < 0) {
+		if ((res = alsa_recover(state, avail)) < 0)
+			return res;
+		if ((res = snd_pcm_htimestamp(state->hndl, &avail, &tstamp)) < 0) {
+			spa_log_warn(state->log, "%s: snd_pcm_htimestamp error: %s",
+				state->props.device, snd_strerror(res));
+			avail = state->threshold * 2;
+		}
+	} else {
+		state->alsa_recovering = false;
+	}
+
+	if ((then = SPA_TIMESPEC_TO_NSEC(&tstamp)) != 0) {
+		if (then < current_time)
+			avail += (current_time - then) * state->rate / SPA_NSEC_PER_SEC;
+		else
+			avail -= (then - current_time) * state->rate / SPA_NSEC_PER_SEC;
+	}
+	return SPA_MIN(avail, state->buffer_frames);
+}
+#endif
+
+static int get_status(struct state *state, uint64_t current_time,
+		snd_pcm_uframes_t *delay, snd_pcm_uframes_t *target)
+{
+	int avail;
+
+	if ((avail = get_avail(state, current_time)) < 0)
+		return avail;
+
+	avail = SPA_MIN(avail, (int)state->buffer_frames);
 
 	*target = state->threshold + state->headroom;
 
@@ -1571,7 +1753,7 @@ static int get_status(struct state *state, snd_pcm_uframes_t *delay, snd_pcm_ufr
 	return 0;
 }
 
-static int update_time(struct state *state, uint64_t nsec, snd_pcm_sframes_t delay,
+static int update_time(struct state *state, uint64_t current_time, snd_pcm_sframes_t delay,
 		snd_pcm_sframes_t target, bool follower)
 {
 	double err, corr;
@@ -1584,8 +1766,8 @@ static int update_time(struct state *state, uint64_t nsec, snd_pcm_sframes_t del
 
 	if (SPA_UNLIKELY(state->dll.bw == 0.0)) {
 		spa_dll_set_bw(&state->dll, SPA_DLL_BW_MAX, state->threshold, state->rate);
-		state->next_time = nsec;
-		state->base_time = nsec;
+		state->next_time = current_time;
+		state->base_time = current_time;
 	}
 	diff = (int32_t) (state->last_threshold - state->threshold);
 
@@ -1594,9 +1776,20 @@ static int update_time(struct state *state, uint64_t nsec, snd_pcm_sframes_t del
 		spa_log_trace(state->log, "%p: follower:%d quantum change %d -> %d (%d) %f",
 				state, follower, state->last_threshold, state->threshold, diff, err);
 		state->last_threshold = state->threshold;
+		state->alsa_sync = true;
 	}
-	err = SPA_CLAMP(err, -state->max_error, state->max_error);
-	corr = spa_dll_update(&state->dll, err);
+	if (err > state->max_error) {
+		err = state->max_error;
+		state->alsa_sync = true;
+	} else if (err < -state->max_error) {
+		err = -state->max_error;
+		state->alsa_sync = true;
+	}
+
+	if (!follower || state->matching)
+		corr = spa_dll_update(&state->dll, err);
+	else
+		corr = 1.0;
 
 	if (diff < 0)
 		state->next_time += diff / corr * 1e9 / state->rate;
@@ -1604,11 +1797,11 @@ static int update_time(struct state *state, uint64_t nsec, snd_pcm_sframes_t del
 	if (SPA_UNLIKELY((state->next_time - state->base_time) > BW_PERIOD)) {
 		state->base_time = state->next_time;
 
-		spa_log_debug(state->log, "%p: follower:%d match:%d rate:%f "
-				"bw:%f thr:%u del:%ld target:%ld err:%f",
-				state, follower, state->matching, corr, state->dll.bw,
-				state->threshold, delay, target,
-				err);
+		spa_log_debug(state->log, "%s: follower:%d match:%d rate:%f "
+				"bw:%f thr:%u del:%ld target:%ld err:%f max:%f",
+				state->props.device, follower, state->matching,
+				corr, state->dll.bw, state->threshold, delay, target,
+				err, state->max_error);
 	}
 
 	if (state->rate_match) {
@@ -1623,7 +1816,7 @@ static int update_time(struct state *state, uint64_t nsec, snd_pcm_sframes_t del
 	state->next_time += state->threshold / corr * 1e9 / state->rate;
 
 	if (SPA_LIKELY(!follower && state->clock)) {
-		state->clock->nsec = nsec;
+		state->clock->nsec = current_time;
 		state->clock->position += state->duration;
 		state->clock->duration = state->duration;
 		state->clock->delay = delay + state->delay;
@@ -1632,7 +1825,7 @@ static int update_time(struct state *state, uint64_t nsec, snd_pcm_sframes_t del
 	}
 
 	spa_log_trace_fp(state->log, "%p: follower:%d %"PRIu64" %f %ld %f %f %u",
-			state, follower, nsec, corr, delay, err, state->threshold * corr,
+			state, follower, current_time, corr, delay, err, state->threshold * corr,
 			state->threshold);
 
 	return 0;
@@ -1671,6 +1864,7 @@ static inline void check_position_config(struct state *state)
 		state->rate_denom = state->position->clock.rate.denom;
 		state->threshold = (state->duration * state->rate + state->rate_denom-1) / state->rate_denom;
 		state->resample = ((uint32_t)state->rate != state->rate_denom) || state->matching;
+		state->alsa_sync = true;
 	}
 }
 
@@ -1678,45 +1872,42 @@ int spa_alsa_write(struct state *state)
 {
 	snd_pcm_t *hndl = state->hndl;
 	const snd_pcm_channel_area_t *my_areas;
-	snd_pcm_uframes_t written, frames, offset, off, to_write, total_written;
+	snd_pcm_uframes_t written, frames, offset, off, to_write, total_written, max_write;
 	snd_pcm_sframes_t commitres;
 	int res = 0;
 
 	check_position_config(state);
 
+	max_write = state->buffer_frames;
+
 	if (state->following && state->alsa_started) {
-		uint64_t nsec;
+		uint64_t current_time;
 		snd_pcm_uframes_t delay, target;
 
-		if (SPA_UNLIKELY((res = get_status(state, &delay, &target)) < 0))
+		current_time = state->position->clock.nsec;
+
+		if (SPA_UNLIKELY((res = get_status(state, current_time, &delay, &target)) < 0))
 			return res;
 
-		if (SPA_UNLIKELY(!state->alsa_recovering && delay > target + state->threshold)) {
+		if (SPA_UNLIKELY(state->alsa_sync)) {
 			spa_log_warn(state->log, "%s: follower delay:%ld target:%ld thr:%u, resync",
 					state->props.device, delay, target, state->threshold);
-			spa_dll_init(&state->dll);
-			state->alsa_sync = true;
-		}
-		if (SPA_UNLIKELY(state->alsa_sync)) {
 			if (delay > target)
 				snd_pcm_rewind(state->hndl, delay - target);
-			else
-				snd_pcm_forward(state->hndl, target - delay);
-
+			else if (delay < target)
+				spa_alsa_silence(state, target - delay);
 			delay = target;
 			state->alsa_sync = false;
 		}
-
-		nsec = state->position->clock.nsec;
-		if (SPA_UNLIKELY((res = update_time(state, nsec, delay, target, true)) < 0))
+		if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, true)) < 0))
 			return res;
 	}
 
 	total_written = 0;
 again:
 
-	frames = state->buffer_frames;
-	if (state->use_mmap) {
+	frames = max_write;
+	if (state->use_mmap && frames > 0) {
 		if (SPA_UNLIKELY((res = snd_pcm_mmap_begin(hndl, &my_areas, &offset, &frames)) < 0)) {
 			spa_log_error(state->log, "%s: snd_pcm_mmap_begin error: %s",
 					state->props.device, snd_strerror(res));
@@ -1806,14 +1997,13 @@ again:
 			state, offset, written, state->sample_count);
 	total_written += written;
 
-	if (state->use_mmap) {
+	if (state->use_mmap && written > 0) {
 		if (SPA_UNLIKELY((commitres = snd_pcm_mmap_commit(hndl, offset, written)) < 0)) {
 			spa_log_error(state->log, "%s: snd_pcm_mmap_commit error: %s",
 					state->props.device, snd_strerror(commitres));
 			if (commitres != -EPIPE && commitres != -ESTRPIPE)
 				return res;
 		}
-
 		if (commitres > 0 && written != (snd_pcm_uframes_t) commitres) {
 			spa_log_warn(state->log, "%s: mmap_commit wrote %ld instead of %ld",
 				     state->props.device, commitres, written);
@@ -1915,62 +2105,47 @@ push_frames(struct state *state,
 int spa_alsa_read(struct state *state)
 {
 	snd_pcm_t *hndl = state->hndl;
-	snd_pcm_uframes_t total_read = 0, to_read;
+	snd_pcm_uframes_t total_read = 0, to_read, max_read;
 	const snd_pcm_channel_area_t *my_areas;
 	snd_pcm_uframes_t read, frames, offset;
 	snd_pcm_sframes_t commitres;
 	int res = 0;
 
-	if (state->position) {
-		check_position_config(state);
+	check_position_config(state);
 
-		if (!state->following) {
-			uint64_t position;
-
-			position = state->position->clock.position;
-			if (state->last_position && state->last_position + state->last_duration != position) {
-				state->alsa_sync = true;
-				spa_log_info(state->log, "%s: discont, resync %"PRIu64" %"PRIu64" %d",
-						state->props.device, state->last_position,
-						position, state->last_duration);
-			}
-			state->last_position = position;
-			state->last_duration = state->duration;
-		}
-	}
+	max_read = state->buffer_frames;
 
 	if (state->following && state->alsa_started) {
-		uint64_t nsec;
-		snd_pcm_uframes_t delay, target;
+		uint64_t current_time;
+		snd_pcm_uframes_t avail, delay, target;
 		uint32_t threshold = state->threshold;
 
-		if ((res = get_status(state, &delay, &target)) < 0)
+		current_time = state->position->clock.nsec;
+
+		if ((res = get_status(state, current_time, &delay, &target)) < 0)
 			return res;
 
-		if (!state->alsa_recovering && (delay < target / 2 || delay > target * 2)) {
+		avail = delay;
+
+		if (state->alsa_sync) {
 			spa_log_warn(state->log, "%s: follower delay:%lu target:%lu thr:%u, resync",
 					state->props.device, delay, target, threshold);
-			spa_dll_init(&state->dll);
-			state->alsa_sync = true;
-		}
-		if (state->alsa_sync) {
-			spa_log_warn(state->log, "%s: follower resync delay:%ld target:%ld thr:%u",
-					state->props.device, delay, target, threshold);
 			if (delay < target)
-				snd_pcm_rewind(state->hndl, target - delay);
+				max_read = target - delay;
 			else if (delay > target)
 				snd_pcm_forward(state->hndl, delay - target);
-
 			delay = target;
 			state->alsa_sync = false;
 		}
 
-		nsec = state->position->clock.nsec;
-		if ((res = update_time(state, nsec, delay, target, true)) < 0)
+		if ((res = update_time(state, current_time, delay, target, true)) < 0)
 			return res;
+
+		if (avail < state->read_size)
+			max_read = 0;
 	}
 
-	frames = state->read_size;
+	frames = SPA_MIN(max_read, state->read_size);
 
 	if (state->use_mmap) {
 		to_read = state->buffer_frames;
@@ -1986,20 +2161,24 @@ int spa_alsa_read(struct state *state)
 		offset = 0;
 	}
 
-	read = push_frames(state, my_areas, offset, frames);
+	if (frames > 0) {
+		read = push_frames(state, my_areas, offset, frames);
+		total_read += read;
+	} else {
+		spa_alsa_skip(state);
+		total_read += state->read_size;
+		read = 0;
+	}
 
-	total_read += read;
-
-	if (state->use_mmap) {
+	if (state->use_mmap && read > 0) {
 		spa_log_trace_fp(state->log, "%p: commit offs:%ld read:%ld count:%"PRIi64, state,
 				offset, read, state->sample_count);
 		if ((commitres = snd_pcm_mmap_commit(hndl, offset, read)) < 0) {
-			spa_log_error(state->log, "%s: snd_pcm_mmap_commit error: %s",
-					state->props.device, snd_strerror(commitres));
+			spa_log_error(state->log, "%s: snd_pcm_mmap_commit error %lu %lu: %s",
+					state->props.device, frames, read, snd_strerror(commitres));
 			if (commitres != -EPIPE && commitres != -ESTRPIPE)
 				return res;
 		}
-
 		if (commitres > 0 && read != (snd_pcm_uframes_t) commitres) {
 			spa_log_warn(state->log, "%s: mmap_commit read %ld instead of %ld",
 				     state->props.device, commitres, read);
@@ -2045,7 +2224,7 @@ int spa_alsa_skip(struct state *state)
 }
 
 
-static int handle_play(struct state *state, uint64_t nsec,
+static int handle_play(struct state *state, uint64_t current_time,
 		snd_pcm_uframes_t delay, snd_pcm_uframes_t target)
 {
 	int res;
@@ -2054,11 +2233,11 @@ static int handle_play(struct state *state, uint64_t nsec,
 		spa_log_trace(state->log, "%p: early wakeup %lu %lu", state, delay, target);
 		if (delay > target * 3)
 			delay = target * 3;
-		state->next_time = nsec + (delay - target) * SPA_NSEC_PER_SEC / state->rate;
+		state->next_time = current_time + (delay - target) * SPA_NSEC_PER_SEC / state->rate;
 		return -EAGAIN;
 	}
 
-	if (SPA_UNLIKELY((res = update_time(state, nsec, delay, target, false)) < 0))
+	if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, false)) < 0))
 		return res;
 
 	if (spa_list_is_empty(&state->ready)) {
@@ -2076,7 +2255,7 @@ static int handle_play(struct state *state, uint64_t nsec,
 	return res;
 }
 
-static int handle_capture(struct state *state, uint64_t nsec,
+static int handle_capture(struct state *state, uint64_t current_time,
 		snd_pcm_uframes_t delay, snd_pcm_uframes_t target)
 {
 	int res;
@@ -2084,12 +2263,12 @@ static int handle_capture(struct state *state, uint64_t nsec,
 
 	if (SPA_UNLIKELY(delay < target)) {
 		spa_log_trace(state->log, "%p: early wakeup %ld %ld", state, delay, target);
-		state->next_time = nsec + (target - delay) * SPA_NSEC_PER_SEC /
+		state->next_time = current_time + (target - delay) * SPA_NSEC_PER_SEC /
 			state->rate;
 		return -EAGAIN;
 	}
 
-	if (SPA_UNLIKELY(res = update_time(state, nsec, delay, target, false)) < 0)
+	if (SPA_UNLIKELY(res = update_time(state, current_time, delay, target, false)) < 0)
 		return res;
 
 	if ((res = spa_alsa_read(state)) < 0)
@@ -2122,17 +2301,20 @@ static void alsa_on_timeout_event(struct spa_source *source)
 {
 	struct state *state = source->data;
 	snd_pcm_uframes_t delay, target;
-	uint64_t expire;
+	uint64_t expire, current_time;
 
 	if (SPA_UNLIKELY(state->started && spa_system_timerfd_read(state->data_system, state->timerfd, &expire) < 0))
 		spa_log_warn(state->log, "%p: error reading timerfd: %m", state);
 
 	check_position_config(state);
 
-	if (SPA_UNLIKELY(get_status(state, &delay, &target) < 0))
-		return;
+	current_time = state->next_time;
 
-	state->current_time = state->next_time;
+	if (SPA_UNLIKELY(get_status(state, current_time, &delay, &target) < 0)) {
+		spa_log_error(state->log, "get_status error");
+		state->next_time += state->threshold * 1e9 / state->rate;
+		goto done;
+	}
 
 #ifndef FASTPATH
 	if (SPA_UNLIKELY(spa_log_level_enabled(state->log, SPA_LOG_LEVEL_TRACE))) {
@@ -2142,16 +2324,24 @@ static void alsa_on_timeout_event(struct spa_source *source)
 		    return;
 		nsec = SPA_TIMESPEC_TO_NSEC(&now);
 		spa_log_trace_fp(state->log, "%p: timeout %lu %lu %"PRIu64" %"PRIu64" %"PRIi64
-				" %d %"PRIi64, state, delay, target, nsec, state->current_time,
-				nsec - state->current_time, state->threshold, state->sample_count);
+				" %d %"PRIi64, state, delay, target, nsec, nsec,
+				nsec - current_time, state->threshold, state->sample_count);
 	}
 #endif
 
 	if (state->stream == SND_PCM_STREAM_PLAYBACK)
-		handle_play(state, state->current_time, delay, target);
+		handle_play(state, current_time, delay, target);
 	else
-		handle_capture(state, state->current_time, delay, target);
+		handle_capture(state, current_time, delay, target);
 
+done:
+	if (state->next_time > current_time + SPA_NSEC_PER_SEC ||
+	    current_time > state->next_time + SPA_NSEC_PER_SEC) {
+		spa_log_error(state->log, "%s: impossible timeout %lu %lu %"PRIu64" %"PRIu64" %"PRIi64
+				" %d %"PRIi64, state->props.device, delay, target, current_time, state->next_time,
+				state->next_time - current_time, state->threshold, state->sample_count);
+		state->next_time = current_time + state->threshold * 1e9 / state->rate;
+	}
 	set_timeout(state, state->next_time);
 }
 
@@ -2205,7 +2395,7 @@ int spa_alsa_start(struct state *state)
 	else {
 		spa_log_warn(state->log, "%s: no position set, using defaults",
 				state->props.device);
-		state->duration = state->props.min_latency;
+		state->duration = 1024;
 		state->rate_denom = state->rate;
 	}
 
