@@ -35,6 +35,11 @@
 #define VOLUME_NORM 1.0f
 
 #include "channelmix-ops.h"
+#include "hilbert.h"
+
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT log_topic
+struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.channelmix");
 
 #define _M(ch)		(1UL << SPA_AUDIO_CHANNEL_ ## ch)
 #define MASK_MONO	_M(FC)|_M(MONO)|_M(UNKNOWN)
@@ -47,8 +52,8 @@
 #define ANY	((uint32_t)-1)
 #define EQ	((uint32_t)-2)
 
-typedef void (*channelmix_func_t) (struct channelmix *mix, uint32_t n_dst, void * SPA_RESTRICT dst[n_dst],
-			uint32_t n_src, const void * SPA_RESTRICT src[n_src], uint32_t n_samples);
+typedef void (*channelmix_func_t) (struct channelmix *mix, void * SPA_RESTRICT dst[],
+			const void * SPA_RESTRICT src[], uint32_t n_samples);
 
 static const struct channelmix_info {
 	uint32_t src_chan;
@@ -80,6 +85,7 @@ static const struct channelmix_info {
 	{ 2, MASK_STEREO, 4, MASK_QUAD, channelmix_f32_2_4_c, 0, "f32_2_4_c" },
 	{ 2, MASK_STEREO, 4, MASK_3_1, channelmix_f32_2_3p1_c, 0, "f32_2_3p1_c" },
 	{ 2, MASK_STEREO, 6, MASK_5_1, channelmix_f32_2_5p1_c, 0, "f32_2_5p1_c" },
+	{ 2, MASK_STEREO, 8, MASK_7_1, channelmix_f32_2_7p1_c, 0, "f32_2_7p1_c" },
 #if defined (HAVE_SSE)
 	{ 6, MASK_5_1, 2, MASK_STEREO, channelmix_f32_5p1_2_sse, SPA_CPU_FLAG_SSE, "f32_5p1_2_sse" },
 #endif
@@ -135,6 +141,7 @@ static const struct channelmix_info *find_channelmix_info(uint32_t src_chan, uin
 
 #define _CH(ch)		((SPA_AUDIO_CHANNEL_ ## ch)-3)
 #define _MASK(ch)	(1ULL << _CH(ch))
+#define FRONT		(_MASK(FC))
 #define STEREO		(_MASK(FL)|_MASK(FR))
 #define REAR		(_MASK(RL)|_MASK(RR))
 #define SIDE		(_MASK(SL)|_MASK(SR))
@@ -144,22 +151,25 @@ static int make_matrix(struct channelmix *mix)
 	float matrix[SPA_AUDIO_MAX_CHANNELS][SPA_AUDIO_MAX_CHANNELS] = {{ 0.0f }};
 	uint64_t src_mask = mix->src_mask;
 	uint64_t dst_mask = mix->dst_mask;
-	uint64_t unassigned;
+	uint64_t unassigned, keep;
 	uint32_t i, j, ic, jc, matrix_encoding = MATRIX_NORMAL;
 	float clev = SQRT1_2;
 	float slev = SQRT1_2;
 	float llev = 0.5f;
 	float maxsum = 0.0f;
-	bool do_upmix = SPA_FLAG_IS_SET(mix->options, CHANNELMIX_OPTION_UPMIX);
 #define _MATRIX(s,d)	matrix[_CH(s)][_CH(d)]
 
-	spa_log_debug(mix->log, "src-mask:%08"PRIx64" dst-mask:%08"PRIx64,
-			src_mask, dst_mask);
+	spa_log_debug(mix->log, "src-mask:%08"PRIx64" dst-mask:%08"PRIx64
+			" options:%08x", src_mask, dst_mask, mix->options);
 
-	/* move the MONO mask to FC so that the lower bits can be shifted
+	/* move the MONO mask to FRONT so that the lower bits can be shifted
 	 * away. */
-	if ((src_mask & (1Ull << SPA_AUDIO_CHANNEL_MONO)) != 0)
-		src_mask |= (1ULL << SPA_AUDIO_CHANNEL_FC);
+	if ((src_mask & (1Ull << SPA_AUDIO_CHANNEL_MONO)) != 0) {
+		if (mix->src_chan == 1)
+			src_mask = 0;
+		else
+			src_mask |= (1ULL << SPA_AUDIO_CHANNEL_FC);
+	}
 	if ((dst_mask & (1Ull << SPA_AUDIO_CHANNEL_MONO)) != 0)
 		dst_mask |= (1ULL << SPA_AUDIO_CHANNEL_FC);
 
@@ -167,13 +177,14 @@ static int make_matrix(struct channelmix *mix)
 	src_mask >>= 3;
 	dst_mask >>= 3;
 
+	/* unknown channels or just 1 channel */
 	if (src_mask == 0 || dst_mask == 0) {
-		if (src_mask == _MASK(FC) && mix->src_chan == 1) {
+		if (mix->src_chan == 1) {
 			/* one FC/MONO src goes everywhere */
 			spa_log_debug(mix->log, "distribute FC/MONO");
 			for (i = 0; i < SPA_AUDIO_MAX_CHANNELS; i++)
 				matrix[i][0]= 1.0f;
-		} else if (dst_mask == _MASK(FC) && mix->dst_chan == 1) {
+		} else if (mix->dst_chan == 1) {
 			/* one FC/MONO dst get average of everything */
 			spa_log_debug(mix->log, "average FC/MONO");
 			for (i = 0; i < SPA_AUDIO_MAX_CHANNELS; i++)
@@ -197,10 +208,21 @@ static int make_matrix(struct channelmix *mix)
 	}
 
 	unassigned = src_mask & ~dst_mask;
+	keep = dst_mask & ~src_mask;
 
-	spa_log_debug(mix->log, "unassigned downmix %08" PRIx64, unassigned);
+	if (!SPA_FLAG_IS_SET(mix->options, CHANNELMIX_OPTION_UPMIX) ||
+	    mix->upmix == CHANNELMIX_UPMIX_NONE)
+		keep = 0;
 
-	if (unassigned & _MASK(FC)){
+	keep |= FRONT;
+	if (mix->lfe_cutoff > 0.0f)
+		keep |= _MASK(LFE);
+	else
+		keep &= ~_MASK(LFE);
+
+	spa_log_debug(mix->log, "unassigned downmix %08" PRIx64 " %08" PRIx64, unassigned, keep);
+
+	if (unassigned & FRONT){
 		if ((dst_mask & STEREO) == STEREO){
 			spa_log_debug(mix->log, "assign FC to STEREO");
 			if(src_mask & STEREO) {
@@ -216,12 +238,13 @@ static int make_matrix(struct channelmix *mix)
 	}
 
 	if (unassigned & STEREO){
-		if (dst_mask & _MASK(FC)) {
+		if (dst_mask & FRONT) {
 			spa_log_debug(mix->log, "assign STEREO to FC");
 			_MATRIX(FC,FL) += SQRT1_2;
 			_MATRIX(FC,FR) += SQRT1_2;
-			if (src_mask & _MASK(FC))
+			if (src_mask & FRONT)
 				_MATRIX(FC,FC) = clev * SQRT2;
+			keep &= ~FRONT;
 		} else {
 			spa_log_warn(mix->log, "can't assign STEREO");
 		}
@@ -251,7 +274,7 @@ static int make_matrix(struct channelmix *mix)
 				_MATRIX(FL,RC) += slev * SQRT1_2;
 				_MATRIX(FR,RC) += slev * SQRT1_2;
 			}
-		} else if (dst_mask & _MASK(FC)) {
+		} else if (dst_mask & FRONT) {
 			spa_log_debug(mix->log, "assign RC to FC");
 			_MATRIX(FC,RC) += slev * SQRT1_2;
 		} else {
@@ -289,7 +312,7 @@ static int make_matrix(struct channelmix *mix)
 				_MATRIX(FL,RL) += slev;
 				_MATRIX(FR,RR) += slev;
 			}
-		} else if (dst_mask & _MASK(FC)) {
+		} else if (dst_mask & FRONT) {
 			spa_log_debug(mix->log, "assign RL+RR to FC");
 			_MATRIX(FC,RL)+= slev * SQRT1_2;
 			_MATRIX(FC,RR)+= slev * SQRT1_2;
@@ -328,7 +351,7 @@ static int make_matrix(struct channelmix *mix)
 				_MATRIX(FL,SL) += slev;
 				_MATRIX(FR,SR) += slev;
 			}
-		} else if (dst_mask & _MASK(FC)) {
+		} else if (dst_mask & FRONT) {
 			spa_log_debug(mix->log, "assign SL+SR to FC");
 			_MATRIX(FC,SL) += slev * SQRT1_2;
 			_MATRIX(FC,SR) += slev * SQRT1_2;
@@ -342,7 +365,7 @@ static int make_matrix(struct channelmix *mix)
 			spa_log_debug(mix->log, "assign FLC+FRC to FL+FR");
 			_MATRIX(FL,FLC)+= 1.0f;
 			_MATRIX(FR,FRC)+= 1.0f;
-		} else if(dst_mask & _MASK(FC)) {
+		} else if(dst_mask & FRONT) {
 			spa_log_debug(mix->log, "assign FLC+FRC to FC");
 			_MATRIX(FC,FLC)+= SQRT1_2;
 			_MATRIX(FC,FRC)+= SQRT1_2;
@@ -352,7 +375,7 @@ static int make_matrix(struct channelmix *mix)
 	}
 	if (unassigned & _MASK(LFE) &&
 	    SPA_FLAG_IS_SET(mix->options, CHANNELMIX_OPTION_MIX_LFE)) {
-		if (dst_mask & _MASK(FC)) {
+		if (dst_mask & FRONT) {
 			spa_log_debug(mix->log, "assign LFE to FC");
 			_MATRIX(FC,LFE) += llev;
 		} else if (dst_mask & STEREO) {
@@ -364,14 +387,12 @@ static int make_matrix(struct channelmix *mix)
 		}
 	}
 
-	if (!do_upmix)
-		goto done;
+	unassigned = dst_mask & ~src_mask & keep;
 
-	unassigned = dst_mask & ~src_mask;
+	spa_log_debug(mix->log, "unassigned upmix %08"PRIx64" lfe:%f",
+			unassigned, mix->lfe_cutoff);
 
-	spa_log_debug(mix->log, "unassigned upmix %08" PRIx64, unassigned);
-
-	if (unassigned & _MASK(FC)) {
+	if (unassigned & FRONT) {
 		if ((src_mask & STEREO) == STEREO) {
 			spa_log_debug(mix->log, "produce FC from STEREO");
 			_MATRIX(FC,FL) += clev;
@@ -380,11 +401,14 @@ static int make_matrix(struct channelmix *mix)
 			spa_log_warn(mix->log, "can't produce FC");
 		}
 	}
-	if (unassigned & _MASK(LFE) && mix->lfe_cutoff > 0.0f) {
+	if (unassigned & _MASK(LFE)) {
 		if ((src_mask & STEREO) == STEREO) {
 			spa_log_debug(mix->log, "produce LFE from STEREO");
 			_MATRIX(LFE,FL) += llev;
 			_MATRIX(LFE,FR) += llev;
+		} else if ((src_mask & FRONT) == FRONT) {
+			spa_log_debug(mix->log, "produce LFE from FC");
+			_MATRIX(LFE,FC) += llev;
 		} else {
 			spa_log_warn(mix->log, "can't produce LFE");
 		}
@@ -394,11 +418,14 @@ static int make_matrix(struct channelmix *mix)
 			spa_log_debug(mix->log, "produce SIDE from REAR");
 			_MATRIX(SL,RL) += 1.0f;
 			_MATRIX(SR,RR) += 1.0f;
-
 		} else if ((src_mask & STEREO) == STEREO) {
 			spa_log_debug(mix->log, "produce SIDE from STEREO");
 			_MATRIX(SL,FL) += 1.0f;
 			_MATRIX(SR,FR) += 1.0f;
+		} else if ((src_mask & FRONT) == FRONT) {
+			spa_log_debug(mix->log, "produce SIDE from FC");
+			_MATRIX(SL,FC) += clev;
+			_MATRIX(SR,FC) += clev;
 		}
 	}
 	if (unassigned & REAR) {
@@ -406,11 +433,14 @@ static int make_matrix(struct channelmix *mix)
 			spa_log_debug(mix->log, "produce REAR from SIDE");
 			_MATRIX(RL,SL) += 1.0f;
 			_MATRIX(RR,SR) += 1.0f;
-
 		} else if ((src_mask & STEREO) == STEREO) {
 			spa_log_debug(mix->log, "produce REAR from STEREO");
 			_MATRIX(RL,FL) += 1.0f;
 			_MATRIX(RR,FR) += 1.0f;
+		} else if ((src_mask & FRONT) == FRONT) {
+			spa_log_debug(mix->log, "produce REAR from FC");
+			_MATRIX(RL,FC) += clev;
+			_MATRIX(RR,FC) += clev;
 		}
 	}
 
@@ -426,21 +456,25 @@ done:
 			sum += fabs(matrix[i][j]);
 		}
 		maxsum = SPA_MAX(maxsum, sum);
-		if (i == _CH(LFE) && do_upmix && mix->lfe_cutoff > 0.0f) {
-			spa_log_debug(mix->log, "channel %d is LFE", ic);
+		if (i == _CH(LFE) && mix->lfe_cutoff > 0.0f) {
+			spa_log_debug(mix->log, "channel %d is LFE cutoff:%f", ic, mix->lfe_cutoff);
 			lr4_set(&mix->lr4[ic], BQ_LOWPASS, mix->lfe_cutoff / mix->freq);
-			mix->lr4_info[ic] = 1;
+		} else if (i == _CH(FC) && mix->fc_cutoff > 0.0f) {
+			spa_log_debug(mix->log, "channel %d is FC cutoff:%f", ic, mix->fc_cutoff);
+			lr4_set(&mix->lr4[ic], BQ_LOWPASS, mix->fc_cutoff / mix->freq);
 		} else {
-			mix->lr4_info[ic] = 0;
+			mix->lr4[ic].active = false;
 		}
 		ic++;
 	}
 	if (SPA_FLAG_IS_SET(mix->options, CHANNELMIX_OPTION_NORMALIZE) &&
 	    maxsum > 1.0f) {
+		spa_log_debug(mix->log, "normalize %f", maxsum);
 		for (i = 0; i < ic; i++)
 			for (j = 0; j < jc; j++)
 		                mix->matrix_orig[i][j] /= maxsum;
 	}
+
 	return 0;
 }
 
@@ -516,11 +550,22 @@ int channelmix_init(struct channelmix *mix)
 	if (info == NULL)
 		return -ENOTSUP;
 
-	spa_log_debug(mix->log, "selected %s", info->name);
-
 	mix->free = impl_channelmix_free;
 	mix->process = info->process;
 	mix->set_volume = impl_channelmix_set_volume;
 	mix->cpu_flags = info->cpu_flags;
+	mix->delay = mix->rear_delay * mix->freq / 1000.0f;
+
+	spa_log_debug(mix->log, "selected %s delay:%d options:%08x", info->name, mix->delay,
+			mix->options);
+
+	if (mix->hilbert_taps > 0) {
+		mix->n_taps = SPA_CLAMP(mix->hilbert_taps, 15u, 255u) | 1;
+		blackman_window(mix->taps, mix->n_taps);
+		hilbert_generate(mix->taps, mix->n_taps);
+	} else {
+		mix->n_taps = 1;
+		mix->taps[0] = 1.0f;
+	}
 	return make_matrix(mix);
 }
