@@ -48,9 +48,9 @@
 #define DEFAULT_RATE		48000
 #define DEFAULT_CHANNELS	2
 
-#define DEFAULT_SAMPLES	8192
 #define MAX_BUFFERS	32
 #define MAX_DATAS	SPA_AUDIO_MAX_CHANNELS
+#define MAX_ALIGN	CHANNELMIX_OPS_MAX_ALIGN
 
 #define DEFAULT_CONTROL_BUFFER_SIZE	32768
 
@@ -82,6 +82,7 @@ struct props {
 	struct volumes soft;
 	struct volumes monitor;
 	unsigned int have_soft_volume:1;
+	unsigned int disabled:1;
 };
 
 static void props_reset(struct props *props)
@@ -142,7 +143,9 @@ struct impl {
 
 	struct spa_log *log;
 	struct spa_cpu *cpu;
+	uint32_t quantum_limit;
 
+	enum spa_direction direction;
 	struct spa_io_position *io_position;
 
 	struct spa_hook_list hooks;
@@ -163,6 +166,7 @@ struct impl {
 	unsigned int started:1;
 	unsigned int is_passthrough:1;
 	uint32_t cpu_flags;
+	uint32_t max_align;
 };
 
 #define IS_CONTROL_PORT(this,d,id)	(id == 1 && d == SPA_DIRECTION_INPUT)
@@ -288,7 +292,8 @@ static void set_volume(struct impl *this)
 {
 	struct volumes *vol;
 
-	if (this->mix.set_volume == NULL)
+	if (this->mix.set_volume == NULL ||
+	    this->props.disabled)
 		return;
 
 	if (this->props.have_soft_volume)
@@ -357,12 +362,14 @@ static int setup_convert(struct impl *this,
 	if ((res = channelmix_init(&this->mix)) < 0)
 		return res;
 
-	remap_volumes(this, src_info);
+	remap_volumes(this, this->direction == SPA_DIRECTION_INPUT ? src_info : dst_info);
 	set_volume(this);
 
 	emit_props_changed(this);
 
 	this->is_passthrough = SPA_FLAG_IS_SET(this->mix.flags, CHANNELMIX_FLAG_IDENTITY);
+	if (!this->is_passthrough && this->props.disabled)
+		return -EINVAL;
 
 	spa_log_debug(this->log, "%p: got channelmix features %08x:%08x flags:%08x passthrough:%d",
 			this, this->cpu_flags, this->mix.cpu_flags,
@@ -460,6 +467,14 @@ static int impl_node_enum_params(void *object, int seq,
 				SPA_PROP_INFO_container, SPA_POD_Id(SPA_TYPE_Array));
 			break;
 		case 8:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_name, SPA_POD_String("channelmix.disable"),
+				SPA_PROP_INFO_description, SPA_POD_String("Disable Channel mixing"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(p->disabled),
+				SPA_PROP_INFO_params, SPA_POD_Bool(true));
+			break;
+		case 9:
 			param = spa_pod_builder_add_object(&b,
 				SPA_TYPE_OBJECT_PropInfo, id,
 				SPA_PROP_INFO_name, SPA_POD_String("channelmix.normalize"),
@@ -468,7 +483,7 @@ static int impl_node_enum_params(void *object, int seq,
 					SPA_FLAG_IS_SET(this->mix.options, CHANNELMIX_OPTION_NORMALIZE)),
 				SPA_PROP_INFO_params, SPA_POD_Bool(true));
 			break;
-		case 9:
+		case 10:
 			param = spa_pod_builder_add_object(&b,
 				SPA_TYPE_OBJECT_PropInfo, id,
 				SPA_PROP_INFO_name, SPA_POD_String("channelmix.mix-lfe"),
@@ -477,7 +492,7 @@ static int impl_node_enum_params(void *object, int seq,
 					SPA_FLAG_IS_SET(this->mix.options, CHANNELMIX_OPTION_MIX_LFE)),
 				SPA_PROP_INFO_params, SPA_POD_Bool(true));
 			break;
-		case 10:
+		case 11:
 			param = spa_pod_builder_add_object(&b,
 				SPA_TYPE_OBJECT_PropInfo, id,
 				SPA_PROP_INFO_name, SPA_POD_String("channelmix.upmix"),
@@ -486,15 +501,76 @@ static int impl_node_enum_params(void *object, int seq,
 					SPA_FLAG_IS_SET(this->mix.options, CHANNELMIX_OPTION_UPMIX)),
 				SPA_PROP_INFO_params, SPA_POD_Bool(true));
 			break;
-		case 11:
+		case 12:
 			param = spa_pod_builder_add_object(&b,
 				SPA_TYPE_OBJECT_PropInfo, id,
 				SPA_PROP_INFO_name, SPA_POD_String("channelmix.lfe-cutoff"),
-				SPA_PROP_INFO_description, SPA_POD_String("LFE cutoff frequency"),
+				SPA_PROP_INFO_description, SPA_POD_String("LFE cutoff frequency (Hz)"),
 				SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Float(
 					this->mix.lfe_cutoff, 0.0, 1000.0),
 				SPA_PROP_INFO_params, SPA_POD_Bool(true));
 			break;
+		case 13:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_name, SPA_POD_String("channelmix.fc-cutoff"),
+				SPA_PROP_INFO_description, SPA_POD_String("FC cutoff frequency (Hz)"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Float(
+					this->mix.fc_cutoff, 0.0, 48000.0),
+				SPA_PROP_INFO_params, SPA_POD_Bool(true));
+			break;
+		case 14:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_name, SPA_POD_String("channelmix.rear-delay"),
+				SPA_PROP_INFO_description, SPA_POD_String("Rear channels delay (ms)"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Float(
+					this->mix.rear_delay, 0.0, 1000.0),
+				SPA_PROP_INFO_params, SPA_POD_Bool(true));
+			break;
+		case 15:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_name, SPA_POD_String("channelmix.stereo-widen"),
+				SPA_PROP_INFO_description, SPA_POD_String("Stereo widen"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Float(
+					this->mix.widen, 0.0, 1.0),
+				SPA_PROP_INFO_params, SPA_POD_Bool(true));
+			break;
+		case 16:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_name, SPA_POD_String("channelmix.hilbert-taps"),
+				SPA_PROP_INFO_description, SPA_POD_String("Taps for phase shift of rear"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Int(
+					this->mix.hilbert_taps, 0, MAX_TAPS),
+				SPA_PROP_INFO_params, SPA_POD_Bool(true));
+			break;
+		case 17:
+		{
+			struct spa_pod_frame f[2];
+			uint32_t i;
+			spa_pod_builder_push_object(&b, &f[0],
+				SPA_TYPE_OBJECT_PropInfo, id);
+			spa_pod_builder_add(&b,
+				SPA_PROP_INFO_name, SPA_POD_String("channelmix.upmix-method"),
+				SPA_PROP_INFO_description, SPA_POD_String("Upmix Method to use"),
+				SPA_PROP_INFO_type, SPA_POD_String(
+					channelmix_upmix_info[this->mix.upmix].label),
+				0);
+			spa_pod_builder_prop(&b, SPA_PROP_INFO_labels, 0);
+			spa_pod_builder_push_struct(&b, &f[1]);
+			for (i = 0; i < SPA_N_ELEMENTS(channelmix_upmix_info); i++) {
+				spa_pod_builder_string(&b, channelmix_upmix_info[i].label);
+				spa_pod_builder_string(&b, channelmix_upmix_info[i].description);
+			}
+			spa_pod_builder_pop(&b, &f[1]);
+			spa_pod_builder_add(&b,
+				SPA_PROP_INFO_params, SPA_POD_Bool(true),
+				0);
+			param = spa_pod_builder_pop(&b, &f[0]);
+			break;
+		}
 		default:
 			return 0;
 		}
@@ -533,6 +609,8 @@ static int impl_node_enum_params(void *object, int seq,
 				0);
 			spa_pod_builder_prop(&b, SPA_PROP_params, 0);
 			spa_pod_builder_push_struct(&b, &f[1]);
+			spa_pod_builder_string(&b, "channelmix.disable");
+			spa_pod_builder_bool(&b, this->props.disabled);
 			spa_pod_builder_string(&b, "channelmix.normalize");
 			spa_pod_builder_bool(&b, SPA_FLAG_IS_SET(this->mix.options,
 						CHANNELMIX_OPTION_NORMALIZE));
@@ -544,6 +622,16 @@ static int impl_node_enum_params(void *object, int seq,
 						CHANNELMIX_OPTION_UPMIX));
 			spa_pod_builder_string(&b, "channelmix.lfe-cutoff");
 			spa_pod_builder_float(&b, this->mix.lfe_cutoff);
+			spa_pod_builder_string(&b, "channelmix.fc-cutoff");
+			spa_pod_builder_float(&b, this->mix.fc_cutoff);
+			spa_pod_builder_string(&b, "channelmix.rear-delay");
+			spa_pod_builder_float(&b, this->mix.rear_delay);
+			spa_pod_builder_string(&b, "channelmix.stereo-widen");
+			spa_pod_builder_float(&b, this->mix.widen);
+			spa_pod_builder_string(&b, "channelmix.hilbert-taps");
+			spa_pod_builder_int(&b, this->mix.hilbert_taps);
+			spa_pod_builder_string(&b, "channelmix.upmix-method");
+			spa_pod_builder_string(&b, channelmix_upmix_info[this->mix.upmix].label);
 			spa_pod_builder_pop(&b, &f[1]);
 			param = spa_pod_builder_pop(&b, &f[0]);
 			break;
@@ -569,21 +657,36 @@ static int impl_node_enum_params(void *object, int seq,
 
 static int channelmix_set_param(struct impl *this, const char *k, const char *s)
 {
-	if (spa_streq(k, "channelmix.normalize"))
+	if (spa_streq(k, "channelmix.disable"))
+		this->props.disabled = spa_atob(s);
+	else if (spa_streq(k, "channelmix.normalize"))
 		SPA_FLAG_UPDATE(this->mix.options, CHANNELMIX_OPTION_NORMALIZE, spa_atob(s));
 	else if (spa_streq(k, "channelmix.mix-lfe"))
 		SPA_FLAG_UPDATE(this->mix.options, CHANNELMIX_OPTION_MIX_LFE, spa_atob(s));
 	else if (spa_streq(k, "channelmix.upmix"))
 		SPA_FLAG_UPDATE(this->mix.options, CHANNELMIX_OPTION_UPMIX, spa_atob(s));
 	else if (spa_streq(k, "channelmix.lfe-cutoff"))
-		this->mix.lfe_cutoff = atoi(s);
-	return 0;
+		spa_atof(s, &this->mix.lfe_cutoff);
+	else if (spa_streq(k, "channelmix.fc-cutoff"))
+		spa_atof(s, &this->mix.fc_cutoff);
+	else if (spa_streq(k, "channelmix.rear-delay"))
+		spa_atof(s, &this->mix.rear_delay);
+	else if (spa_streq(k, "channelmix.stereo-widen"))
+		spa_atof(s, &this->mix.widen);
+	else if (spa_streq(k, "channelmix.hilbert-taps"))
+		spa_atou32(s, &this->mix.hilbert_taps, 0);
+	else if (spa_streq(k, "channelmix.upmix-method"))
+		this->mix.upmix = channelmix_upmix_from_label(s);
+	else
+		return 0;
+	return 1;
 }
 
 static int parse_prop_params(struct impl *this, struct spa_pod *params)
 {
 	struct spa_pod_parser prs;
 	struct spa_pod_frame f;
+	int changed = 0;
 
 	spa_pod_parser_pod(&prs, params);
 	if (spa_pod_parser_push_struct(&prs, &f) < 0)
@@ -592,7 +695,7 @@ static int parse_prop_params(struct impl *this, struct spa_pod *params)
 	while (true) {
 		const char *name;
 		struct spa_pod *pod;
-		char value[512];
+		char value[512], buf[128];
 
 		if (spa_pod_parser_get_string(&prs, &name) < 0)
 			break;
@@ -603,8 +706,9 @@ static int parse_prop_params(struct impl *this, struct spa_pod *params)
 		if (spa_pod_is_string(pod)) {
 			spa_pod_copy_string(pod, sizeof(value), value);
 		} else if (spa_pod_is_float(pod)) {
-			snprintf(value, sizeof(value), "%f",
-					SPA_POD_VALUE(struct spa_pod_float, pod));
+			snprintf(value, sizeof(value), "%s",
+					spa_json_format_float(buf, sizeof(buf),
+						SPA_POD_VALUE(struct spa_pod_float, pod)));
 		} else if (spa_pod_is_int(pod)) {
 			snprintf(value, sizeof(value), "%d",
 					SPA_POD_VALUE(struct spa_pod_int, pod));
@@ -616,9 +720,11 @@ static int parse_prop_params(struct impl *this, struct spa_pod *params)
 			continue;
 
 		spa_log_info(this->log, "key:'%s' val:'%s'", name, value);
-		channelmix_set_param(this, name, value);
+		changed += channelmix_set_param(this, name, value);
 	}
-	return 0;
+	if (changed && !this->props.disabled)
+		channelmix_init(&this->mix);
+	return changed;
 }
 
 static int apply_props(struct impl *this, const struct spa_pod *param)
@@ -693,8 +799,9 @@ static int apply_props(struct impl *this, const struct spa_pod *param)
 			break;
 		}
 	}
+
 	if (changed) {
-		struct port *port = GET_IN_PORT(this, 0);
+		struct port *port = GET_PORT(this, this->direction, 0);
 		if (have_soft_volume)
 			p->have_soft_volume = true;
 		else if (have_channel_volume)
@@ -702,6 +809,7 @@ static int apply_props(struct impl *this, const struct spa_pod *param)
 
 		if (port->have_format)
 			remap_volumes(this, &port->format);
+
 		set_volume(this);
 	}
 	return changed;
@@ -863,6 +971,7 @@ static int port_enum_formats(void *object,
 		} else {
 			struct spa_pod_frame f;
 			struct port *other;
+			int32_t channels, min = 1, max = INT32_MAX;
 
 			other = GET_PORT(this, SPA_DIRECTION_REVERSE(direction), 0);
 
@@ -875,19 +984,24 @@ static int port_enum_formats(void *object,
 				0);
 
 			if (other->have_format) {
+				channels = other->format.info.raw.channels;
+				if (this->props.disabled)
+					min = max = channels;
+
 				spa_pod_builder_add(builder,
 					SPA_FORMAT_AUDIO_rate, SPA_POD_Int(other->format.info.raw.rate),
 					SPA_FORMAT_AUDIO_channels, SPA_POD_CHOICE_RANGE_Int(
-						other->format.info.raw.channels, 1, INT32_MAX),
+						channels, min, max),
 					0);
 			} else {
 				uint32_t rate = this->io_position ?
 					this->io_position->clock.rate.denom : DEFAULT_RATE;
+				channels = DEFAULT_CHANNELS;
 
 				spa_pod_builder_add(builder,
 					SPA_FORMAT_AUDIO_rate, SPA_POD_CHOICE_RANGE_Int(rate, 0, INT32_MAX),
 					SPA_FORMAT_AUDIO_channels, SPA_POD_CHOICE_RANGE_Int(
-						DEFAULT_CHANNELS, 1, INT32_MAX),
+						channels, min, max),
 					0);
 			}
 			*param = spa_pod_builder_pop(builder, &f);
@@ -978,7 +1092,7 @@ impl_node_port_enum_params(void *object, int seq,
 				size = other->size / other->stride;
 			} else {
 				buffers = 1;
-				size = DEFAULT_SAMPLES;
+				size = this->quantum_limit;
 			}
 
 			param = spa_pod_builder_add_object(&b,
@@ -1180,7 +1294,7 @@ impl_node_port_use_buffers(void *object,
 					      buffers[i]);
 				return -EINVAL;
 			}
-			if (!SPA_IS_ALIGNED(d[j].data, 16)) {
+			if (!SPA_IS_ALIGNED(d[j].data, this->max_align)) {
 				spa_log_warn(this->log, "%p: memory %d on buffer %d not aligned",
 						this, j, i);
 			}
@@ -1263,8 +1377,7 @@ static int impl_node_port_reuse_buffer(void *object, uint32_t port_id, uint32_t
 }
 
 static int channelmix_process_control(struct impl *this, struct port *ctrlport,
-				      uint32_t n_dst, void * SPA_RESTRICT dst[n_dst],
-				      uint32_t n_src, const void * SPA_RESTRICT src[n_src],
+				      void * SPA_RESTRICT dst[], const void * SPA_RESTRICT src[],
 				      uint32_t n_samples)
 {
 	struct spa_pod_control *c, *prev = NULL;
@@ -1307,10 +1420,10 @@ static int channelmix_process_control(struct impl *this, struct port *ctrlport,
 		spa_log_trace_fp(this->log, "%p: process %d %d", this,
 				c->offset, chunk);
 
-		channelmix_process(&this->mix, n_dst, dst, n_src, src, chunk);
-		for (i = 0; i < n_src; i++)
+		channelmix_process(&this->mix, dst, src, chunk);
+		for (i = 0; i < this->mix.src_chan; i++)
 			s[i] += chunk;
-		for (i = 0; i < n_dst; i++)
+		for (i = 0; i < this->mix.dst_chan; i++)
 			d[i] += chunk;
 
 		avail_samples -= chunk;
@@ -1323,7 +1436,7 @@ static int channelmix_process_control(struct impl *this, struct port *ctrlport,
 	 * remaining samples */
 	spa_log_trace_fp(this->log, "%p: remain %d", this, avail_samples);
 	if (avail_samples > 0)
-		channelmix_process(&this->mix, n_dst, dst, n_src, src, avail_samples);
+		channelmix_process(&this->mix, dst, src, avail_samples);
 
 	return 1;
 }
@@ -1415,14 +1528,14 @@ static int impl_node_process(void *object)
 		if (!is_passthrough) {
 			if (ctrlport->ctrl != NULL) {
 				/* if return value is 1, the sequence has been processed */
-				if (channelmix_process_control(this, ctrlport, n_dst_datas, dst_datas,
-						n_src_datas, src_datas, n_samples) == 1) {
+				if (channelmix_process_control(this, ctrlport, dst_datas,
+						src_datas, n_samples) == 1) {
 					ctrlio->status = SPA_STATUS_OK;
 					ctrlport->ctrl = NULL;
 				}
 			} else {
-				channelmix_process(&this->mix, n_dst_datas, dst_datas,
-						n_src_datas, src_datas, n_samples);
+				channelmix_process(&this->mix, dst_datas,
+						src_datas, n_samples);
 			}
 		}
 	}
@@ -1532,18 +1645,36 @@ impl_init(const struct spa_handle_factory *factory,
 	spa_log_topic_init(this->log, log_topic);
 
 	this->cpu = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_CPU);
-	if (this->cpu)
+	if (this->cpu) {
 		this->cpu_flags = spa_cpu_get_flags(this->cpu);
+		this->max_align = SPA_MIN(MAX_ALIGN, spa_cpu_get_max_align(this->cpu));
+	}
 
 	spa_hook_list_init(&this->hooks);
 
 	props_reset(&this->props);
 
+	this->mix.options = CHANNELMIX_OPTION_UPMIX;
+	this->mix.upmix = CHANNELMIX_UPMIX_SIMPLE;
+	this->mix.log = this->log;
+	this->mix.lfe_cutoff = 120.0f;
+	this->mix.fc_cutoff = 6000.0f;
+	this->mix.rear_delay = 12.0f;
+	this->mix.widen = 0.1f;
+
 	for (i = 0; info && i < info->n_items; i++) {
 		const char *k = info->items[i].key;
 		const char *s = info->items[i].value;
 		if (spa_streq(k, SPA_KEY_AUDIO_POSITION))
 			this->props.n_channels = parse_position(this->props.channel_map, s, strlen(s));
+		else if (spa_streq(k, "clock.quantum-limit"))
+			spa_atou32(s, &this->quantum_limit, 0);
+		else if (spa_streq(k, "factory.mode")) {
+			if (spa_streq(s, "merge"))
+				this->direction = SPA_DIRECTION_OUTPUT;
+			else
+				this->direction = SPA_DIRECTION_INPUT;
+		}
 		else
 			channelmix_set_param(this, k, s);
 
