@@ -60,9 +60,6 @@ extern "C" {
 
 #define PIPEWIRE_BATTERY_PROVIDER "/org/freedesktop/pipewire/battery"
 
-#define MIN_LATENCY	512
-#define MAX_LATENCY	1024
-
 #define OBJECT_MANAGER_INTROSPECT_XML                                          \
 	DBUS_INTROSPECT_1_0_XML_DOCTYPE_DECL_NODE                                  \
 	"<node>\n"                                                                 \
@@ -332,11 +329,13 @@ struct spa_bt_adapter {
 	uint32_t bluetooth_class;
 	uint32_t profiles;
 	int powered;
+	unsigned int has_msbc:1;
+	unsigned int msbc_probed:1;
 	unsigned int endpoints_registered:1;
 	unsigned int application_registered:1;
 	unsigned int player_registered:1;
-	unsigned int has_battery_provider;
-	unsigned int battery_provider_unavailable;
+	unsigned int has_battery_provider:1;
+	unsigned int battery_provider_unavailable:1;
 };
 
 enum spa_bt_form_factor {
@@ -473,6 +472,8 @@ struct spa_bt_device {
 	 * under same device. So it's safe to cache activation info here. */
 	bool a2dp_volume_active[2];
 
+	uint64_t last_bluez_action_time;
+
 	struct spa_hook_list listener_list;
 	bool added;
 
@@ -495,6 +496,7 @@ int spa_bt_device_ensure_hfp_codec(struct spa_bt_device *device, unsigned int co
 int spa_bt_device_supports_hfp_codec(struct spa_bt_device *device, unsigned int codec);
 int spa_bt_device_release_transports(struct spa_bt_device *device);
 int spa_bt_device_report_battery_level(struct spa_bt_device *device, uint8_t percentage);
+void spa_bt_device_update_last_bluez_action_time(struct spa_bt_device *device);
 
 #define spa_bt_device_emit(d,m,v,...)			spa_hook_list_call(&(d)->listener_list, \
 								struct spa_bt_device_events,	\
@@ -579,6 +581,8 @@ struct spa_bt_transport {
 	struct spa_bt_transport_volume volumes[SPA_BT_VOLUME_ID_TERM];
 
 	int acquire_refcount;
+	bool acquired;
+	bool keepalive;
 	int fd;
 	uint16_t read_mtu;
 	uint16_t write_mtu;
@@ -608,6 +612,7 @@ bool spa_bt_transport_volume_enabled(struct spa_bt_transport *transport);
 
 int spa_bt_transport_acquire(struct spa_bt_transport *t, bool optional);
 int spa_bt_transport_release(struct spa_bt_transport *t);
+int spa_bt_transport_keepalive(struct spa_bt_transport *t, bool keepalive);
 int spa_bt_transport_ensure_sco_io(struct spa_bt_transport *t, struct spa_loop *data_loop);
 
 #define spa_bt_transport_emit(t,m,v,...)		spa_hook_list_call(&(t)->listener_list, \
@@ -682,6 +687,7 @@ enum spa_bt_feature {
 	SPA_BT_FEATURE_HW_VOLUME_MIC	= (1 << 4),
 	SPA_BT_FEATURE_SBC_XQ		= (1 << 5),
 	SPA_BT_FEATURE_FASTSTREAM	= (1 << 6),
+	SPA_BT_FEATURE_A2DP_DUPLEX	= (1 << 7),
 };
 
 struct spa_bt_quirks;
@@ -693,6 +699,8 @@ int spa_bt_quirks_get_features(const struct spa_bt_quirks *quirks,
 		uint32_t *features);
 void spa_bt_quirks_destroy(struct spa_bt_quirks *quirks);
 
+int spa_bt_adapter_has_msbc(struct spa_bt_adapter *adapter);
+
 struct spa_bt_backend_implementation {
 #define SPA_VERSION_BT_BACKEND_IMPLEMENTATION	0
 	uint32_t version;
