@@ -43,10 +43,14 @@
 #include <spa/param/audio/raw.h>
 #include <spa/param/profiler.h>
 #include <spa/pod/builder.h>
+#include <spa/support/plugin.h>
 #include <spa/utils/json.h>
+#include <spa/utils/names.h>
 #include <spa/utils/result.h>
 #include <spa/utils/ringbuffer.h>
 #include <spa/utils/string.h>
+#include <spa/support/plugin-loader.h>
+#include <spa/interfaces/audio/aec.h>
 
 #include <pipewire/private.h>
 #include <pipewire/impl.h>
@@ -54,8 +58,6 @@
 
 #include <pipewire/extensions/profiler.h>
 
-#include "module-echo-cancel/echo-cancel.h"
-
 /** \page page_module_echo_cancel PipeWire Module: Echo Cancel
  *
  * The `echo-cancel` module performs echo cancellation. The module creates
@@ -68,8 +70,8 @@
  *
  * - `source.props = {}`: properties to be passed to the source stream
  * - `sink.props = {}`: properties to be passed to the sink stream
- * - `aec.method = <str>`: the echo cancellation method. Currently supported:
- * `webrtc`. Leave unset to use the default method (`webrtc`).
+ * - `library.name = <str>`: the echo cancellation library  Currently supported:
+ * `aec/libspa-aec-webrtc`. Leave unset to use the default method (`aec/libspa-aec-webrtc`).
  * - `aec.args = <str>`: arguments to pass to the echo cancellation method
  *
  * ## General options
@@ -94,7 +96,7 @@
  * context.modules = [
  *  {   name = libpipewire-module-echo-cancel
  *      args = {
- *          # aec.method = webrtc
+ *          # library.name  = aec/libspa-aec-webrtc
  *          # node.latency = 1024/48000
  *          source.props = {
  *             node.name = "Echo Cancellation Source"
@@ -126,6 +128,7 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 /* Hopefully this is enough for any combination of AEC engine and resampler
  * input requirement for rate matching */
 #define MAX_BUFSIZE_MS 100
+#define DELAY_MS 0
 
 static const struct spa_dict_item module_props[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Wim Taymans <wim.taymans@gmail.com>" },
@@ -135,7 +138,9 @@ static const struct spa_dict_item module_props[] = {
 				"[ audio.rate=<sample rate> ] "
 				"[ audio.channels=<number of channels> ] "
 				"[ audio.position=<channel map> ] "
-				"[ aec.method=<aec method> ] "
+				"[ buffer.max_size=<max buffer size in ms> ] "
+				"[ buffer.play_delay=<delay as fraction> ] "
+				"[ library.name =<library name> ] "
 				"[ aec.args=<aec arguments> ] "
 				"[ source.props=<properties> ] "
 				"[ sink.props=<properties> ] " },
@@ -146,7 +151,6 @@ struct impl {
 	struct pw_context *context;
 
 	struct pw_impl_module *module;
-	struct pw_work_queue *work;
 	struct spa_hook module_listener;
 
 	uint32_t id;
@@ -175,35 +179,27 @@ struct impl {
 	void *play_buffer[SPA_AUDIO_MAX_CHANNELS];
 	uint32_t play_ringsize;
 	struct spa_ringbuffer play_ring;
+	struct spa_ringbuffer play_delayed_ring;
 	struct spa_io_rate_match *play_rate_match;
 
 	void *out_buffer[SPA_AUDIO_MAX_CHANNELS];
 	uint32_t out_ringsize;
 	struct spa_ringbuffer out_ring;
 
-	const struct echo_cancel_info *aec_info;
-	void *aec;
+	struct spa_audio_aec *aec;
 	uint32_t aec_blocksize;
 
 	unsigned int capture_ready:1;
 	unsigned int sink_ready:1;
 
 	unsigned int do_disconnect:1;
-	unsigned int unloading:1;
-};
 
-static void do_unload_module(void *obj, void *data, int res, uint32_t id)
-{
-	struct impl *impl = data;
-	pw_impl_module_destroy(impl->module);
-}
-static void unload_module(struct impl *impl)
-{
-	if (!impl->unloading) {
-		impl->unloading = true;
-		pw_work_queue_add(impl->work, impl, 0, do_unload_module, impl);
-	}
-}
+	uint32_t max_buffer_size;
+	uint32_t buffer_delay;
+
+	struct spa_handle *spa_handle;
+	struct spa_plugin_loader *loader;
+};
 
 static void process(struct impl *impl)
 {
@@ -211,13 +207,15 @@ static void process(struct impl *impl)
 	struct pw_buffer *pout;
 	float rec_buf[impl->info.channels][impl->aec_blocksize / sizeof(float)];
 	float play_buf[impl->info.channels][impl->aec_blocksize / sizeof(float)];
+	float play_delayed_buf[impl->info.channels][impl->aec_blocksize / sizeof(float)];
 	float out_buf[impl->info.channels][impl->aec_blocksize / sizeof(float)];
 	const float *rec[impl->info.channels];
 	const float *play[impl->info.channels];
+	const float *play_delayed[impl->info.channels];
 	float *out[impl->info.channels];
 	struct spa_data *dd;
 	uint32_t i, size;
-	uint32_t rindex, pindex, oindex, avail;
+	uint32_t rindex, pindex, oindex, pdindex, avail;
 	int32_t stride = 0;
 
 	if ((pout = pw_stream_dequeue_buffer(impl->playback)) == NULL) {
@@ -231,12 +229,15 @@ static void process(struct impl *impl)
 
 	spa_ringbuffer_get_read_index(&impl->rec_ring, &rindex);
 	spa_ringbuffer_get_read_index(&impl->play_ring, &pindex);
+	spa_ringbuffer_get_read_index(&impl->play_delayed_ring, &pdindex);
 
 	for (i = 0; i < impl->info.channels; i++) {
 		/* captured samples, with echo from sink */
 		rec[i] = &rec_buf[i][0];
 		/* echo from sink */
 		play[i] = &play_buf[i][0];
+		/* echo from sink delayed */
+		play_delayed[i] = &play_delayed_buf[i][0];
 		/* filtered samples, without echo from sink */
 		out[i] = &out_buf[i][0];
 
@@ -250,6 +251,11 @@ static void process(struct impl *impl)
 				impl->play_ringsize, pindex % impl->play_ringsize,
 				(void *)play[i], size);
 
+		stride = 0;
+		spa_ringbuffer_read_data(&impl->play_delayed_ring, impl->play_buffer[i],
+				impl->play_ringsize, pdindex % impl->play_ringsize,
+				(void *)play_delayed[i], size);
+
 		/* output to sink, just copy */
 		dd = &pout->buffer->datas[i];
 		memcpy(dd->data, play[i], size);
@@ -261,13 +267,12 @@ static void process(struct impl *impl)
 
 	spa_ringbuffer_read_update(&impl->rec_ring, rindex + size);
 	spa_ringbuffer_read_update(&impl->play_ring, pindex + size);
+	spa_ringbuffer_read_update(&impl->play_delayed_ring, pdindex + size);
 
 	pw_stream_queue_buffer(impl->playback, pout);
 
-	pw_stream_trigger_process(impl->playback);
-
 	/* Now run the canceller */
-	echo_cancel_run(impl->aec_info, impl->aec, rec,	play, out, size / sizeof(float));
+	spa_audio_aec_run(impl->aec, rec, play_delayed, out, size / sizeof(float));
 
 	/* Next, copy over the output to the output ringbuffer */
 	avail = spa_ringbuffer_get_write_index(&impl->out_ring, &oindex);
@@ -320,7 +325,6 @@ static void process(struct impl *impl)
 		spa_ringbuffer_read_update(&impl->out_ring, oindex);
 		avail -= size;
 	}
-	pw_stream_trigger_process(impl->source);
 
 done:
 	impl->sink_ready = false;
@@ -556,6 +560,9 @@ static void sink_process(void *data)
 		spa_ringbuffer_get_read_index(&impl->play_ring, &rindex);
 		spa_ringbuffer_read_update(&impl->play_ring, rindex + drop);
 
+		spa_ringbuffer_get_read_index(&impl->play_delayed_ring, &rindex);
+		spa_ringbuffer_read_update(&impl->play_delayed_ring, rindex + drop);
+
 		avail += drop;
 	}
 
@@ -616,6 +623,7 @@ static int setup_streams(struct impl *impl)
 	struct spa_pod_builder b;
 	struct pw_properties *props;
 	const char *str;
+	uint32_t index;
 
 	props = pw_properties_new(
 			PW_KEY_NODE_NAME, "echo-cancel-capture",
@@ -628,8 +636,8 @@ static int setup_streams(struct impl *impl)
 		pw_properties_set(props, PW_KEY_NODE_LINK_GROUP, str);
 	if ((str = pw_properties_get(impl->source_props, PW_KEY_NODE_LATENCY)) != NULL)
 		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
-	else if (impl->aec_info->latency)
-		pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec_info->latency);
+	else if (impl->aec->latency)
+		pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec->latency);
 
 	impl->capture = pw_stream_new(impl->core,
 			"Echo-Cancel Capture", props);
@@ -661,8 +669,8 @@ static int setup_streams(struct impl *impl)
 		pw_properties_set(props, PW_KEY_NODE_LINK_GROUP, str);
 	if ((str = pw_properties_get(impl->sink_props, PW_KEY_NODE_LATENCY)) != NULL)
 		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
-	else if (impl->aec_info->latency)
-		pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec_info->latency);
+	else if (impl->aec->latency)
+		pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec->latency);
 
 	impl->playback = pw_stream_new(impl->core,
 			"Echo-Cancel Playback", props);
@@ -701,8 +709,7 @@ static int setup_streams(struct impl *impl)
 			PW_DIRECTION_OUTPUT,
 			PW_ID_ANY,
 			PW_STREAM_FLAG_MAP_BUFFERS |
-			PW_STREAM_FLAG_RT_PROCESS |
-			PW_STREAM_FLAG_TRIGGER,
+			PW_STREAM_FLAG_RT_PROCESS,
 			params, n_params)) < 0)
 		return res;
 
@@ -719,14 +726,13 @@ static int setup_streams(struct impl *impl)
 			PW_ID_ANY,
 			PW_STREAM_FLAG_AUTOCONNECT |
 			PW_STREAM_FLAG_MAP_BUFFERS |
-			PW_STREAM_FLAG_RT_PROCESS |
-			PW_STREAM_FLAG_TRIGGER,
+			PW_STREAM_FLAG_RT_PROCESS,
 			params, n_params)) < 0)
 		return res;
 
-	impl->rec_ringsize = sizeof(float) * MAX_BUFSIZE_MS * impl->info.rate / 1000;
-	impl->play_ringsize = sizeof(float) * MAX_BUFSIZE_MS * impl->info.rate / 1000;
-	impl->out_ringsize = sizeof(float) * MAX_BUFSIZE_MS * impl->info.rate / 1000;
+	impl->rec_ringsize = sizeof(float) * impl->max_buffer_size * impl->info.rate / 1000;
+	impl->play_ringsize = sizeof(float) * ((impl->max_buffer_size * impl->info.rate / 1000) + impl->buffer_delay);
+	impl->out_ringsize = sizeof(float) * impl->max_buffer_size * impl->info.rate / 1000;
 	for (i = 0; i < impl->info.channels; i++) {
 		impl->rec_buffer[i] = malloc(impl->rec_ringsize);
 		impl->play_buffer[i] = malloc(impl->play_ringsize);
@@ -734,8 +740,14 @@ static int setup_streams(struct impl *impl)
 	}
 	spa_ringbuffer_init(&impl->rec_ring);
 	spa_ringbuffer_init(&impl->play_ring);
+	spa_ringbuffer_init(&impl->play_delayed_ring);
 	spa_ringbuffer_init(&impl->out_ring);
 
+	spa_ringbuffer_get_write_index(&impl->play_ring, &index);
+	spa_ringbuffer_write_update(&impl->play_ring, index + (sizeof(float) * (impl->buffer_delay)));
+	spa_ringbuffer_get_read_index(&impl->play_ring, &index);
+	spa_ringbuffer_read_update(&impl->play_ring, index + (sizeof(float) * (impl->buffer_delay)));
+
 	return 0;
 }
 
@@ -747,7 +759,7 @@ static void core_error(void *data, uint32_t id, int seq, int res, const char *me
 			id, seq, res, spa_strerror(res), message);
 
 	if (id == PW_ID_CORE && res == -EPIPE)
-		unload_module(impl);
+		pw_impl_module_schedule_destroy(impl->module);
 }
 
 static const struct pw_core_events core_events = {
@@ -760,7 +772,7 @@ static void core_destroy(void *d)
 	struct impl *impl = d;
 	spa_hook_remove(&impl->core_listener);
 	impl->core = NULL;
-	unload_module(impl);
+	pw_impl_module_schedule_destroy(impl->module);
 }
 
 static const struct pw_proxy_events core_proxy_events = {
@@ -780,14 +792,11 @@ static void impl_destroy(struct impl *impl)
 		pw_stream_destroy(impl->sink);
 	if (impl->core && impl->do_disconnect)
 		pw_core_disconnect(impl->core);
-	if (impl->aec)
-		echo_cancel_destroy(impl->aec_info, impl->aec);
+	if (impl->spa_handle)
+		spa_plugin_loader_unload(impl->loader, impl->spa_handle);
 	pw_properties_free(impl->source_props);
 	pw_properties_free(impl->sink_props);
 
-	if (impl->work)
-		pw_work_queue_cancel(impl->work, impl, SPA_ID_INVALID);
-
 	for (i = 0; i < impl->info.channels; i++) {
 		if (impl->rec_buffer[i])
 			free(impl->rec_buffer[i]);
@@ -803,7 +812,6 @@ static void impl_destroy(struct impl *impl)
 static void module_destroy(void *data)
 {
 	struct impl *impl = data;
-	impl->unloading = true;
 	spa_hook_remove(&impl->module_listener);
 	impl_destroy(impl);
 }
@@ -870,7 +878,10 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	struct impl *impl;
 	uint32_t id = pw_global_get_id(pw_impl_module_get_global(module));
 	const char *str;
-	int res;
+	const char *path;
+	int res = 0;
+	struct spa_handle *handle = NULL;
+	void *iface;
 
 	PW_LOG_TOPIC_INIT(mod_topic);
 
@@ -901,12 +912,6 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	impl->id = id;
 	impl->module = module;
 	impl->context = context;
-	impl->work = pw_context_get_work_queue(context);
-	if (impl->work == NULL) {
-		res = -errno;
-		pw_log_error( "can't create work queue: %m");
-		goto error;
-	}
 
 	if (pw_properties_get(props, PW_KEY_NODE_GROUP) == NULL)
 		pw_properties_setf(props, PW_KEY_NODE_GROUP, "echo-cancel-%u", id);
@@ -944,31 +949,68 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	if (pw_properties_get(impl->sink_props, PW_KEY_MEDIA_CLASS) == NULL)
 		pw_properties_set(impl->sink_props, PW_KEY_MEDIA_CLASS, "Audio/Sink");
 
-	if ((str = pw_properties_get(props, "aec.method")) == NULL)
-		str = "webrtc";
+	if ((str = pw_properties_get(props, "aec.method")) != NULL)
+		pw_log_warn("aec.method is not supported anymore use library.name");
 
-#ifdef HAVE_WEBRTC
-	if (spa_streq(str, "webrtc"))
-		impl->aec_info = echo_cancel_webrtc;
-	else
-#endif
-		impl->aec_info = echo_cancel_null;
+	/* Use webrtc as default */
+	if ((path = pw_properties_get(props, "library.name")) == NULL)
+		path = "aec/libspa-aec-webrtc";
+
+	struct spa_dict_item info_items[] = {
+		{ SPA_KEY_LIBRARY_NAME, path },
+	};
+	struct spa_dict info = SPA_DICT_INIT_ARRAY(info_items);
+
+	impl->loader = spa_support_find(context->support, context->n_support, SPA_TYPE_INTERFACE_PluginLoader);
+	if (impl->loader == NULL) {
+		pw_log_error("a plugin loader is needed");
+		return -EINVAL;
+	}
+
+	handle = spa_plugin_loader_load(impl->loader, SPA_NAME_AEC, &info);
+	if (handle == NULL) {
+		pw_log_error("AEC codec plugin %s not available library.name %s", SPA_NAME_AEC, path);
+		return -ENOENT;
+	}
+
+	if ((res = spa_handle_get_interface(handle, SPA_TYPE_INTERFACE_AUDIO_AEC, &iface)) < 0) {
+		pw_log_error("can't get %s interface %d", SPA_TYPE_INTERFACE_AUDIO_AEC, res);
+		return res;
+	}
+	impl->aec = iface;
+	impl->spa_handle = handle;
+	if (impl->aec->iface.version != SPA_VERSION_AUDIO_AEC) {
+		pw_log_error("codec plugin %s has incompatible ABI version (%d != %d)",
+			SPA_NAME_AEC, impl->aec->iface.version, SPA_VERSION_AUDIO_AEC);
+		res = -ENOENT;
+		goto error;
+	}
+
+	(void)SPA_SUPPORT_INIT(SPA_TYPE_INTERFACE_AUDIO_AEC, (struct spa_audio_aec *)impl->aec);
+
+	pw_log_info("Using plugin AEC %s", impl->aec->name);
 
 	if ((str = pw_properties_get(props, "aec.args")) != NULL)
 		aec_props = pw_properties_new_string(str);
 	else
 		aec_props = pw_properties_new(NULL, NULL);
 
-	impl->aec = echo_cancel_create(impl->aec_info, aec_props, &impl->info);
+	res = spa_audio_aec_init(impl->aec, &aec_props->dict, &impl->info);
 
 	pw_properties_free(aec_props);
 
-	if (impl->aec_info->latency) {
+	if (res < 0) {
+		pw_log_error("codec plugin %s create failed: %s", impl->aec->name,
+				spa_strerror(res));
+		goto error;
+	}
+
+	if (impl->aec->latency) {
 		unsigned int num, denom, req_num, req_denom;
 		unsigned int factor = 0;
 		unsigned int new_num = 0;
 
-		spa_assert_se(sscanf(impl->aec_info->latency, "%u/%u", &num, &denom) == 2);
+		spa_assert_se(sscanf(impl->aec->latency, "%u/%u", &num, &denom) == 2);
 
 		if ((str = pw_properties_get(props, PW_KEY_NODE_LATENCY)) != NULL) {
 			sscanf(str, "%u/%u", &req_num, &req_denom);
@@ -977,8 +1019,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 		}
 
 		if (factor == 0 || new_num == 0) {
-			pw_log_info("Setting node latency to %s", impl->aec_info->latency);
-			pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec_info->latency);
+			pw_log_info("Setting node latency to %s", impl->aec->latency);
+			pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec->latency);
 			impl->aec_blocksize = sizeof(float) * impl->info.rate * num / denom;
 		} else {
 			pw_log_info("Setting node latency to %u/%u", new_num, req_denom);
@@ -1011,6 +1053,25 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	copy_props(impl, props, PW_KEY_NODE_VIRTUAL);
 	copy_props(impl, props, PW_KEY_NODE_LATENCY);
 
+	impl->max_buffer_size = pw_properties_get_uint32(props,"buffer.max_size", MAX_BUFSIZE_MS);
+
+	if ((str = pw_properties_get(props, "buffer.play_delay")) != NULL) {
+		int req_num, req_denom;
+		if (sscanf(str, "%u/%u", &req_num, &req_denom) == 2) {
+			if (req_denom != 0) {
+				impl->buffer_delay = (impl->info.rate*req_num)/req_denom;
+			} else {
+				impl->buffer_delay = DELAY_MS * impl->info.rate / 1000;
+				pw_log_warn("Sample rate for buffer.play_delay is 0 using default");
+			}
+		} else {
+			impl->buffer_delay = DELAY_MS * impl->info.rate / 1000;
+			pw_log_warn("Wrong value/format for buffer.play_delay using default");
+		}
+	} else {
+		impl->buffer_delay = DELAY_MS * impl->info.rate / 1000;
+	}
+
 	pw_properties_free(props);
 
 	pw_proxy_add_listener((struct pw_proxy*)impl->core,
