@@ -28,6 +28,7 @@
 #include <stdio.h>
 #include <limits.h>
 
+#include <spa/support/plugin.h>
 #include <spa/utils/defs.h>
 #include <spa/utils/list.h>
 #include <spa/utils/string.h>
@@ -46,7 +47,9 @@ struct fc_port {
 #define FC_PORT_AUDIO	(1ULL << 3)
 	uint64_t flags;
 
+#define FC_HINT_BOOLEAN		(1ULL << 2)
 #define FC_HINT_SAMPLE_RATE	(1ULL << 3)
+#define FC_HINT_INTEGER		(1ULL << 5)
 	uint64_t hint;
 	float def;
 	float min;
@@ -93,8 +96,9 @@ static inline void fc_descriptor_free(struct fc_descriptor *desc)
 		desc->free(desc);
 }
 
-struct fc_plugin *load_ladspa_plugin(const char *path, const char *config);
-struct fc_plugin *load_lv2_plugin(const char *path, const char *config);
-struct fc_plugin *load_builtin_plugin(const char *path, const char *config);
-
-void init_builtin_plugin(uint32_t cpu_flags);
+struct fc_plugin *load_ladspa_plugin(const struct spa_support *support, uint32_t n_support,
+		const char *path, const char *config);
+struct fc_plugin *load_lv2_plugin(const struct spa_support *support, uint32_t n_support,
+		const char *path, const char *config);
+struct fc_plugin *load_builtin_plugin(const struct spa_support *support, uint32_t n_support,
+		const char *path, const char *config);
