@@ -67,25 +67,18 @@ static int module_roc_source_load(struct client *client, struct module *module)
 {
 	struct module_roc_source_data *data = module->user_data;
 	FILE *f;
-	const char *str;
 	char *args;
 	size_t size;
 
-	f = open_memstream(&args, &size);
+	pw_properties_setf(data->source_props, "pulse.module.id",
+			"%u", module->index);
+
+	if ((f = open_memstream(&args, &size)) == NULL)
+		return -errno;
+
 	fprintf(f, "{");
-	/* Can't just serialise this dict because the "null" method gets
-	 * interpreted as a JSON null */
-	if ((str = pw_properties_get(data->roc_props, "local.ip")))
-		fprintf(f, " local.ip = \"%s\"", str);
-	if ((str = pw_properties_get(data->roc_props, "local.source.port")))
-		fprintf(f, " local.source.port = \"%s\"", str);
-	if ((str = pw_properties_get(data->roc_props, "local.repair.port")))
-		fprintf(f, " local.repair.port = \"%s\"", str);
-	if ((str = pw_properties_get(data->roc_props, "sess.latency.msec")))
-		fprintf(f, " sess.latency.msec = \"%s\"", str);
-	if ((str = pw_properties_get(data->roc_props, "resampler.profile")))
-		fprintf(f, " resampler.profile = \"%s\"", str);
-	fprintf(f, " } source.props = {");
+	pw_properties_serialize_dict(f, &data->roc_props->dict, 0);
+	fprintf(f, " source.props = {");
 	pw_properties_serialize_dict(f, &data->source_props->dict, 0);
 	fprintf(f, " } }");
 	fclose(f);
@@ -132,6 +125,7 @@ static const struct spa_dict_item module_roc_source_info[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Sanchayan Maity <sanchayan@asymptotic.io>" },
 	{ PW_KEY_MODULE_DESCRIPTION, "roc source" },
 	{ PW_KEY_MODULE_USAGE, "source_name=<name for the source> "
+				"source_properties=<properties for the source> "
 				"resampler_profile=<empty>|disable|high|medium|low "
 				"sess_latency_msec=<target network latency in milliseconds> "
 				"local_ip=<local receiver ip> "
@@ -165,6 +159,10 @@ struct module *create_module_roc_source(struct impl *impl, const char *argument)
 		pw_properties_set(source_props, PW_KEY_NODE_NAME, str);
 		pw_properties_set(props, "source_name", NULL);
 	}
+	if ((str = pw_properties_get(props, "source_properties")) != NULL) {
+		module_args_add_props(source_props, str);
+		pw_properties_set(props, "source_properties", NULL);
+	}
 
 	if ((str = pw_properties_get(props, PW_KEY_MEDIA_CLASS)) == NULL) {
 		pw_properties_set(props, PW_KEY_MEDIA_CLASS, "Audio/Source");
