@@ -34,6 +34,7 @@
 #include <unistd.h>
 #include <assert.h>
 #include <ctype.h>
+#include <locale.h>
 
 #include <sndfile.h>
 
@@ -52,6 +53,7 @@
 #include <pipewire/extensions/metadata.h>
 
 #include "midifile.h"
+#include "dsffile.h"
 
 #define DEFAULT_MEDIA_TYPE	"Audio"
 #define DEFAULT_MIDI_MEDIA_TYPE	"Midi"
@@ -89,6 +91,10 @@ typedef int (*fill_fn)(struct data *d, void *dest, unsigned int n_frames);
 struct target {
 	struct spa_list link;
 	uint32_t id;
+#define TARGET_TYPE_SINK	0
+#define TARGET_TYPE_SOURCE	1
+#define TARGET_TYPE_STREAM	2
+	uint32_t type;
 	char *name;
 	char *desc;
 	int prio;
@@ -116,7 +122,10 @@ struct data {
 
 	enum mode mode;
 	bool verbose;
-	bool is_midi;
+#define TYPE_PCM	0
+#define TYPE_MIDI	1
+#define TYPE_DSD	2
+	int data_type;
 	const char *remote_name;
 	const char *media_type;
 	const char *media_category;
@@ -161,6 +170,11 @@ struct data {
 		struct midi_file *file;
 		struct midi_file_info info;
 	} midi;
+	struct {
+		struct dsf_file *file;
+		struct dsf_file_info info;
+		struct dsf_layout layout;
+	} dsf;
 };
 
 static inline int
@@ -610,7 +624,7 @@ target_destroy(struct target *target)
 }
 
 static struct target *
-target_create(uint32_t id, const char *name, const char *desc, int prio)
+target_create(uint32_t id, uint32_t type, const char *name, const char *desc, int prio)
 {
 	struct target *target;
 
@@ -618,6 +632,7 @@ target_create(uint32_t id, const char *name, const char *desc, int prio)
 	if (!target)
 		return NULL;
 	target->id = id;
+	target->type = type;
 	target->name = strdup(name);
 	target->desc = strdup(desc ? : "");
 	target->prio = prio;
@@ -634,7 +649,7 @@ static void on_core_info(void *userdata, const struct pw_core_info *info)
 	struct data *data = userdata;
 
 	if (data->verbose)
-		fprintf(stdout, "remote %"PRIu32" is named \"%s\"\n",
+		printf("remote %"PRIu32" is named \"%s\"\n",
 				info->id, info->name);
 }
 
@@ -680,7 +695,7 @@ static int json_object_find(const char *obj, const char *key, char *value, size_
 	if (spa_json_enter_object(&it[0], &it[1]) <= 0)
 		return -EINVAL;
 
-	while (spa_json_get_string(&it[1], k, sizeof(k)-1) > 0) {
+	while (spa_json_get_string(&it[1], k, sizeof(k)) > 0) {
 		if (spa_streq(k, key)) {
 			if (spa_json_get_string(&it[1], value, len) <= 0)
 				continue;
@@ -728,8 +743,8 @@ static void registry_event_global(void *userdata, uint32_t id,
 	const struct spa_dict_item *item;
 	const char *name, *desc, *media_class, *prio_session;
 	int prio;
-	enum mode mode = mode_none;
 	struct target *target;
+	uint32_t ttype;
 
 	/* only once */
 	if (data->targets_listed)
@@ -742,6 +757,9 @@ static void registry_event_global(void *userdata, uint32_t id,
 	if (spa_streq(type, PW_TYPE_INTERFACE_Metadata)) {
 		if (data->metadata != NULL)
 			return;
+		if ((name = spa_dict_lookup(props, PW_KEY_METADATA_NAME)) != NULL &&
+		    !spa_streq(name, "default"))
+			return;
 
 		data->metadata = pw_registry_bind(data->registry,
 				id, type, PW_VERSION_METADATA, 0);
@@ -761,15 +779,27 @@ static void registry_event_global(void *userdata, uint32_t id,
 		if (!name || !media_class)
 			return;
 
-		/* get allowed mode from the media class */
-		/* TODO extend to something else besides Audio/Source|Sink */
-		if (spa_streq(media_class, "Audio/Source"))
-			mode = mode_record;
-		else if (spa_streq(media_class, "Audio/Sink"))
-			mode = mode_playback;
+		if (desc == NULL)
+			desc = name;
 
-		/* modes must match */
-		if (mode != data->mode)
+		/* get allowed mode from the media class */
+		if (spa_streq(media_class, "Audio/Source")) {
+			if (data->mode != mode_record)
+				return;
+			ttype = TARGET_TYPE_SOURCE;
+		}
+		else if (spa_streq(media_class, "Stream/Output/Audio")) {
+			if (data->mode != mode_record)
+				return;
+			ttype = TARGET_TYPE_STREAM;
+		}
+		else if (spa_streq(media_class, "Audio/Sink")) {
+			if (data->mode != mode_playback &&
+			    data->mode != mode_record)
+				return;
+			ttype = TARGET_TYPE_SINK;
+		}
+		else
 			return;
 
 		prio = prio_session ? atoi(prio_session) : -1;
@@ -779,11 +809,11 @@ static void registry_event_global(void *userdata, uint32_t id,
 					id, type, name, media_class, desc ? : "", prio);
 
 			spa_dict_for_each(item, props) {
-				fprintf(stdout, "\t\t%s = \"%s\"\n", item->key, item->value);
+				printf("\t\t%s = \"%s\"\n", item->key, item->value);
 			}
 		}
 
-		target = target_create(id, name, desc, prio);
+		target = target_create(id, ttype, name, desc, prio);
 		if (target)
 			spa_list_append(&data->targets, &target->link);
 	}
@@ -851,7 +881,8 @@ on_io_changed(void *userdata, uint32_t id, void *data, uint32_t size)
 		d->position = data;
 		break;
 	case SPA_IO_RateMatch:
-		d->rate_match = data;
+		if (d->data_type == TYPE_PCM)
+			d->rate_match = data;
 		break;
 	default:
 		break;
@@ -862,8 +893,37 @@ static void
 on_param_changed(void *userdata, uint32_t id, const struct spa_pod *param)
 {
 	struct data *data = userdata;
+	struct spa_audio_info info = { 0 };
+	int err;
+
 	if (data->verbose)
 		printf("stream param change: id=%"PRIu32"\n", id);
+
+	if (id != SPA_PARAM_Format || param == NULL)
+		return;
+
+	if ((err = spa_format_parse(param, &info.media_type, &info.media_subtype)) < 0)
+		return;
+
+	if (info.media_type != SPA_MEDIA_TYPE_audio ||
+	    info.media_subtype != SPA_MEDIA_SUBTYPE_dsd)
+		return;
+
+	if (spa_format_audio_dsd_parse(param, &info.info.dsd) < 0)
+		return;
+
+	data->dsf.layout.interleave = info.info.dsd.interleave,
+	data->dsf.layout.channels = info.info.dsd.channels;
+	data->dsf.layout.lsb = info.info.dsd.bitorder == SPA_PARAM_BITORDER_lsb;
+
+	data->stride = data->dsf.layout.channels * SPA_ABS(data->dsf.layout.interleave);
+
+	if (data->verbose) {
+		printf("DSD out: channels:%d bitorder:%s interleave:%d\n",
+				data->dsf.layout.channels,
+				data->dsf.layout.lsb ? "lsb" : "msb",
+				data->dsf.layout.interleave);
+	}
 }
 
 static void on_process(void *userdata)
@@ -1058,6 +1118,7 @@ static void show_usage(const char *name, bool is_error)
 		   _("  -p, --playback                        Playback mode\n"
 		     "  -r, --record                          Recording mode\n"
 		     "  -m, --midi                            Midi mode\n"
+		     "  -d, --dsd                             DSD mode\n"
 		     "\n"), fp);
 	}
 }
@@ -1177,6 +1238,48 @@ static int setup_midifile(struct data *data)
 	return 0;
 }
 
+struct dsd_layout_info {
+	 uint32_t type;
+	 struct spa_audio_layout_info info;
+};
+static const struct dsd_layout_info dsd_layouts[] = {
+	{ 1, { SPA_AUDIO_LAYOUT_Mono, }, },
+	{ 2, { SPA_AUDIO_LAYOUT_Stereo, }, },
+	{ 3, { SPA_AUDIO_LAYOUT_2FC }, },
+	{ 4, { SPA_AUDIO_LAYOUT_Quad }, },
+	{ 5, { SPA_AUDIO_LAYOUT_3_1 }, },
+	{ 6, { SPA_AUDIO_LAYOUT_5_0R }, },
+	{ 7, { SPA_AUDIO_LAYOUT_5_1R }, },
+};
+
+static int dsf_play(struct data *d, void *src, unsigned int n_frames)
+{
+	return dsf_file_read(d->dsf.file, src, n_frames, &d->dsf.layout);
+}
+
+static int setup_dsffile(struct data *data)
+{
+	if (data->mode == mode_record)
+		return -ENOTSUP;
+
+	data->dsf.file = dsf_file_open(data->filename, "r", &data->dsf.info);
+	if (data->dsf.file == NULL) {
+		fprintf(stderr, "error: can't read dsf file '%s': %m\n", data->filename);
+		return -errno;
+	}
+
+	if (data->verbose)
+		printf("opened file \"%s\" channels:%d rate:%d samples:%"PRIu64" bitorder:%s\n",
+				data->filename,
+				data->dsf.info.channels, data->dsf.info.rate,
+				data->dsf.info.samples,
+				data->dsf.info.lsb ? "lsb" : "msb");
+
+	data->fill = dsf_play;
+
+	return 0;
+}
+
 static int fill_properties(struct data *data)
 {
 	static const char * const table[] = {
@@ -1387,6 +1490,7 @@ int main(int argc, char *argv[])
 	int exit_code = EXIT_FAILURE, c, ret;
 	enum pw_stream_flags flags = 0;
 
+	setlocale(LC_ALL, "");
 	pw_init(&argc, &argv);
 
 	flags |= PW_STREAM_FLAG_AUTOCONNECT;
@@ -1398,16 +1502,21 @@ int main(int argc, char *argv[])
 		prog = argv[0];
 
 	/* prime the mode from the program name */
-	if (spa_streq(prog, "pw-play"))
+	if (spa_streq(prog, "pw-play")) {
 		data.mode = mode_playback;
-	else if (spa_streq(prog, "pw-record"))
+		data.data_type = TYPE_PCM;
+	} else if (spa_streq(prog, "pw-record")) {
 		data.mode = mode_record;
-	else if (spa_streq(prog, "pw-midiplay")) {
+		data.data_type = TYPE_PCM;
+	} else if (spa_streq(prog, "pw-midiplay")) {
 		data.mode = mode_playback;
-		data.is_midi = true;
+		data.data_type = TYPE_MIDI;
 	} else if (spa_streq(prog, "pw-midirecord")) {
 		data.mode = mode_record;
-		data.is_midi = true;
+		data.data_type = TYPE_MIDI;
+	} else if (spa_streq(prog, "pw-dsdplay")) {
+		data.mode = mode_playback;
+		data.data_type = TYPE_DSD;
 	} else
 		data.mode = mode_none;
 
@@ -1418,7 +1527,7 @@ int main(int argc, char *argv[])
 	/* initialize list every time */
 	spa_list_init(&data.targets);
 
-	while ((c = getopt_long(argc, argv, "hvprmR:q:", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "hvprmdR:q:", long_options, NULL)) != -1) {
 
 		switch (c) {
 
@@ -1427,7 +1536,7 @@ int main(int argc, char *argv[])
 			return EXIT_SUCCESS;
 
 		case OPT_VERSION:
-			fprintf(stdout, "%s\n"
+			printf("%s\n"
 				"Compiled with libpipewire %s\n"
 				"Linked with libpipewire %s\n",
 				prog,
@@ -1448,7 +1557,11 @@ int main(int argc, char *argv[])
 			break;
 
 		case 'm':
-			data.is_midi = true;
+			data.data_type = TYPE_MIDI;
+			break;
+
+		case 'd':
+			data.data_type = TYPE_DSD;
 			break;
 
 		case 'R':
@@ -1538,10 +1651,14 @@ int main(int argc, char *argv[])
 	}
 
 	if (!data.media_type) {
-		if (data.is_midi)
+		switch (data.data_type) {
+		case TYPE_MIDI:
 			data.media_type = DEFAULT_MIDI_MEDIA_TYPE;
-		else
+			break;
+		default:
 			data.media_type = DEFAULT_MEDIA_TYPE;
+			break;
+		}
 	}
 	if (!data.media_category)
 		data.media_category = data.mode == mode_playback ?
@@ -1637,12 +1754,21 @@ int main(int argc, char *argv[])
 	data.sync = pw_core_sync(data.core, 0, data.sync);
 
 	if (!data.list_targets) {
-		struct spa_audio_info_raw info;
 
-		if (data.is_midi)
-			ret = setup_midifile(&data);
-		else
+		switch (data.data_type) {
+		case TYPE_PCM:
 			ret = setup_sndfile(&data);
+			break;
+		case TYPE_MIDI:
+			ret = setup_midifile(&data);
+			break;
+		case TYPE_DSD:
+			ret = setup_dsffile(&data);
+			break;
+		default:
+			ret = -ENOTSUP;
+			break;
+		}
 
 		if (ret < 0) {
 			fprintf(stderr, "error: open failed: %s\n", spa_strerror(ret));
@@ -1654,8 +1780,10 @@ int main(int argc, char *argv[])
 				goto error_usage;
 			}
 		}
-
-		if (!data.is_midi) {
+		switch (data.data_type) {
+		case TYPE_PCM:
+		{
+			struct spa_audio_info_raw info;
 			info = SPA_AUDIO_INFO_RAW_INIT(
 				.flags = data.channelmap.n_channels ? 0 : SPA_AUDIO_FLAG_UNPOSITIONED,
 				.format = data.spa_format,
@@ -1666,13 +1794,35 @@ int main(int argc, char *argv[])
 				memcpy(info.position, data.channelmap.channels, data.channels * sizeof(int));
 
 			params[0] = spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &info);
-		} else {
+			break;
+		}
+		case TYPE_MIDI:
 			params[0] = spa_pod_builder_add_object(&b,
 					SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
 					SPA_FORMAT_mediaType,		SPA_POD_Id(SPA_MEDIA_TYPE_application),
 					SPA_FORMAT_mediaSubtype,	SPA_POD_Id(SPA_MEDIA_SUBTYPE_control));
 
 			pw_properties_set(data.props, PW_KEY_FORMAT_DSP, "8 bit raw midi");
+			break;
+		case TYPE_DSD:
+		{
+			struct spa_audio_info_dsd info;
+			size_t i;
+
+			spa_zero(info);
+			info.channels = data.dsf.info.channels;
+			info.rate = data.dsf.info.rate / 8;
+
+			for (i = 0; i < SPA_N_ELEMENTS(dsd_layouts); i++) {
+				if (dsd_layouts[i].type != data.dsf.info.channel_type)
+					continue;
+				info.channels = dsd_layouts[i].info.n_channels;
+				memcpy(info.position, dsd_layouts[i].info.position,
+						info.channels * sizeof(uint32_t));
+			}
+			params[0] = spa_format_audio_dsd_build(&b, SPA_PARAM_EnumFormat, &info);
+			break;
+		}
 		}
 
 		data.stream = pw_stream_new(data.core, prog, data.props);
@@ -1750,7 +1900,24 @@ int main(int argc, char *argv[])
 			}
 			printf("Available targets (\"*\" denotes default): %s\n", default_name);
 			spa_list_for_each(target, &data.targets, link) {
-				printf("%s\t%"PRIu32": description=\"%s\" prio=%d\n",
+				if (target->type != TARGET_TYPE_SOURCE)
+					continue;
+				printf("%s\t%"PRIu32": source description=\"%s\" prio=%d\n",
+				       target == target_default ? "*" : "",
+				       target->id, target->desc, target->prio);
+			}
+			spa_list_for_each(target, &data.targets, link) {
+				if (target->type != TARGET_TYPE_SINK)
+					continue;
+				printf("%s\t%"PRIu32": %s description=\"%s\" prio=%d\n",
+					target == target_default ? "*" : "",
+					target->id, data.mode == mode_record ? "monitor" : "sink",
+					target->desc, target->prio);
+			}
+			spa_list_for_each(target, &data.targets, link) {
+				if (target->type != TARGET_TYPE_STREAM)
+					continue;
+				printf("%s\t%"PRIu32": stream description=\"%s\" prio=%d\n",
 				       target == target_default ? "*" : "",
 				       target->id, target->desc, target->prio);
 			}
@@ -1766,14 +1933,21 @@ int main(int argc, char *argv[])
 	}
 
 error_connect_fail:
-	if (data.stream)
+	if (data.stream) {
+		spa_hook_remove(&data.stream_listener);
 		pw_stream_destroy(data.stream);
+	}
 error_no_stream:
-	if (data.metadata)
+	if (data.metadata) {
+		spa_hook_remove(&data.metadata_listener);
 		pw_proxy_destroy((struct pw_proxy*)data.metadata);
-	if (data.registry)
+	}
+	if (data.registry) {
+		spa_hook_remove(&data.registry_listener);
 		pw_proxy_destroy((struct pw_proxy*)data.registry);
+	}
 error_no_registry:
+	spa_hook_remove(&data.core_listener);
 	pw_core_disconnect(data.core);
 error_ctx_connect_failed:
 	pw_context_destroy(data.context);
