@@ -25,8 +25,12 @@
 #include <stdio.h>
 #include <signal.h>
 #include <getopt.h>
+#include <unistd.h>
+#include <locale.h>
 
 #include <spa/utils/result.h>
+#include <spa/utils/string.h>
+#include <spa/utils/ansi.h>
 #include <spa/debug/pod.h>
 #include <spa/debug/format.h>
 #include <spa/debug/types.h>
@@ -37,6 +41,20 @@ struct proxy_data;
 
 typedef void (*print_func_t) (struct proxy_data *data);
 
+static struct pprefix {
+	const char *prefix;
+	const char *suffix;
+} pprefix[2] = {
+	{ .prefix = " ", .suffix = "" },
+	{ .prefix = "*", .suffix = "" },
+};
+
+#define with_prefix(use_prefix_) \
+   for (bool once_ = !!printf("%s", (pprefix[!!(use_prefix_)]).prefix); \
+	once_; \
+	once_ = false, printf("%s", (pprefix[!!(use_prefix_)]).suffix))
+
+
 struct param {
 	struct spa_list link;
 	uint32_t id;
@@ -56,6 +74,7 @@ struct data {
 	struct spa_hook registry_listener;
 
 	struct spa_list pending_list;
+	struct spa_list global_list;
 };
 
 struct proxy_data {
@@ -71,6 +90,7 @@ struct proxy_data {
 	struct spa_hook proxy_listener;
 	struct spa_hook object_listener;
 	int pending_seq;
+	struct spa_list global_link;
 	struct spa_list pending_link;
 	print_func_t print_func;
 	struct spa_list param_list;
@@ -146,47 +166,57 @@ static void event_param(void *object, int seq, uint32_t id,
 
 	p->id = id;
 	p->seq = seq;
-	p->param = SPA_MEMBER(p, sizeof(struct param), struct spa_pod);
+	p->param = SPA_PTROFF(p, sizeof(struct param), struct spa_pod);
 	p->changed = true;
 	memcpy(p->param, param, SPA_POD_SIZE(param));
 	spa_list_append(&data->param_list, &p->link);
 }
 
-static void print_params(struct proxy_data *data, char mark)
+static void print_params(struct proxy_data *data, bool use_prefix)
 {
 	struct param *p;
 
-	printf("%c\tparams:\n", mark);
+	with_prefix(use_prefix) {
+		printf("\tparams:\n");
+	}
+
 	spa_list_for_each(p, &data->param_list, link) {
-		printf("%c\t  id:%u (%s)\n", p->changed ? mark : ' ', p->id,
-			spa_debug_type_find_name(spa_type_param, p->id));
-		if (spa_pod_is_object_type(p->param, SPA_TYPE_OBJECT_Format))
-			spa_debug_format(10, NULL, p->param);
-		else
-			spa_debug_pod(10, NULL, p->param);
+		with_prefix(p->changed) {
+			printf("\t  id:%u (%s)\n",
+				p->id,
+				spa_debug_type_find_name(spa_type_param, p->id));
+			if (spa_pod_is_object_type(p->param, SPA_TYPE_OBJECT_Format))
+				spa_debug_format(10, NULL, p->param);
+			else
+				spa_debug_pod(10, NULL, p->param);
+		}
 		p->changed = false;
 	}
 }
 
-static void print_properties(const struct spa_dict *props, char mark)
+static void print_properties(const struct spa_dict *props, bool use_prefix)
 {
 	const struct spa_dict_item *item;
 
-	printf("%c\tproperties:\n", mark);
-	if (props == NULL || props->n_items == 0) {
-		printf("\t\tnone\n");
-		return;
+	with_prefix(use_prefix) {
+		printf("\tproperties:\n");
+		if (props == NULL || props->n_items == 0) {
+			printf("\t\tnone\n");
+			return;
+		}
 	}
 
 	spa_dict_for_each(item, props) {
-		if (item->value)
-			printf("%c\t\t%s = \"%s\"\n", mark, item->key, item->value);
-		else
-			printf("%c\t\t%s = (null)\n", mark, item->key);
+		with_prefix(use_prefix) {
+			if (item->value)
+				printf("\t\t%s = \"%s\"\n", item->key, item->value);
+			else
+				printf("\t\t%s = (null)\n", item->key);
+		}
 	}
 }
 
-#define MARK_CHANGE(f) ((print_mark && ((info)->change_mask & (f))) ? '*' : ' ')
+#define MARK_CHANGE(f) (!!(print_mark && ((info)->change_mask & (f))))
 
 static void on_core_info(void *data, const struct pw_core_info *info)
 {
@@ -259,12 +289,18 @@ static void print_node(struct proxy_data *data)
 	printf("\ttype: %s (version %d)\n", data->type, data->version);
 	if (print_all) {
 		print_params(data, MARK_CHANGE(PW_NODE_CHANGE_MASK_PARAMS));
-		printf("%c\tinput ports: %u/%u\n", MARK_CHANGE(PW_NODE_CHANGE_MASK_INPUT_PORTS),
+		with_prefix(MARK_CHANGE(PW_NODE_CHANGE_MASK_INPUT_PORTS)) {
+			printf("\tinput ports: %u/%u\n",
 				info->n_input_ports, info->max_input_ports);
-		printf("%c\toutput ports: %u/%u\n", MARK_CHANGE(PW_NODE_CHANGE_MASK_OUTPUT_PORTS),
+		}
+		with_prefix(MARK_CHANGE(PW_NODE_CHANGE_MASK_OUTPUT_PORTS)) {
+			printf("\toutput ports: %u/%u\n",
 				info->n_output_ports, info->max_output_ports);
-		printf("%c\tstate: \"%s\"", MARK_CHANGE(PW_NODE_CHANGE_MASK_STATE),
+		}
+		with_prefix(MARK_CHANGE(PW_NODE_CHANGE_MASK_STATE)) {
+			printf("\tstate: \"%s\"",
 				pw_node_state_as_string(info->state));
+		}
 		if (info->state == PW_NODE_STATE_ERROR && info->error)
 			printf(" \"%s\"\n", info->error);
 		else
@@ -456,17 +492,21 @@ static void link_event_info(void *object, const struct pw_link_info *info)
 	printf("\tinput-node-id: %u\n", info->input_node_id);
 	printf("\tinput-port-id: %u\n", info->input_port_id);
 	if (print_all) {
-		printf("%c\tstate: \"%s\"", MARK_CHANGE(PW_LINK_CHANGE_MASK_STATE),
+		with_prefix(MARK_CHANGE(PW_LINK_CHANGE_MASK_STATE)) {
+			printf("\tstate: \"%s\"",
 				pw_link_state_as_string(info->state));
+		}
 		if (info->state == PW_LINK_STATE_ERROR && info->error)
 			printf(" \"%s\"\n", info->error);
 		else
 			printf("\n");
-		printf("%c\tformat:\n", MARK_CHANGE(PW_LINK_CHANGE_MASK_FORMAT));
-		if (info->format)
-			spa_debug_format(2, NULL, info->format);
-		else
-			printf("\t\tnone\n");
+		with_prefix(MARK_CHANGE(PW_LINK_CHANGE_MASK_FORMAT)) {
+			printf("\tformat:\n");
+			if (info->format)
+				spa_debug_format(2, NULL, info->format);
+			else
+				printf("\t\tnone\n");
+		}
 		print_properties(info->props, MARK_CHANGE(PW_LINK_CHANGE_MASK_PROPS));
 	}
 }
@@ -544,18 +584,22 @@ removed_proxy (void *data)
 static void
 destroy_proxy (void *data)
 {
-        struct proxy_data *pd = data;
+	struct proxy_data *pd = data;
+
+	spa_list_remove(&pd->global_link);
+
+	spa_hook_remove(&pd->object_listener);
+	spa_hook_remove(&pd->proxy_listener);
 
 	clear_params(pd);
 	remove_pending(pd);
+	free(pd->type);
 
-        if (pd->info == NULL)
-                return;
-
+	if (pd->info == NULL)
+		return;
 	if (pd->destroy)
 		pd->destroy(pd->info);
-        pd->info = NULL;
-	free(pd->type);
+	pd->info = NULL;
 }
 
 static const struct pw_proxy_events proxy_events = {
@@ -568,42 +612,42 @@ static void registry_event_global(void *data, uint32_t id,
 				  uint32_t permissions, const char *type, uint32_t version,
 				  const struct spa_dict *props)
 {
-        struct data *d = data;
-        struct pw_proxy *proxy;
-        uint32_t client_version;
-        const void *events;
+	struct data *d = data;
+	struct pw_proxy *proxy;
+	uint32_t client_version;
+	const void *events;
 	struct proxy_data *pd;
 	pw_destroy_t destroy;
 	print_func_t print_func = NULL;
 
-	if (strcmp(type, PW_TYPE_INTERFACE_Node) == 0) {
+	if (spa_streq(type, PW_TYPE_INTERFACE_Node)) {
 		events = &node_events;
 		client_version = PW_VERSION_NODE;
 		destroy = (pw_destroy_t) pw_node_info_free;
 		print_func = print_node;
-	} else if (strcmp(type, PW_TYPE_INTERFACE_Port) == 0) {
+	} else if (spa_streq(type, PW_TYPE_INTERFACE_Port)) {
 		events = &port_events;
 		client_version = PW_VERSION_PORT;
 		destroy = (pw_destroy_t) pw_port_info_free;
 		print_func = print_port;
-	} else if (strcmp(type, PW_TYPE_INTERFACE_Module) == 0) {
+	} else if (spa_streq(type, PW_TYPE_INTERFACE_Module)) {
 		events = &module_events;
 		client_version = PW_VERSION_MODULE;
 		destroy = (pw_destroy_t) pw_module_info_free;
-	} else if (strcmp(type, PW_TYPE_INTERFACE_Device) == 0) {
+	} else if (spa_streq(type, PW_TYPE_INTERFACE_Device)) {
 		events = &device_events;
 		client_version = PW_VERSION_DEVICE;
 		destroy = (pw_destroy_t) pw_device_info_free;
 		print_func = print_device;
-	} else if (strcmp(type, PW_TYPE_INTERFACE_Factory) == 0) {
+	} else if (spa_streq(type, PW_TYPE_INTERFACE_Factory)) {
 		events = &factory_events;
 		client_version = PW_VERSION_FACTORY;
 		destroy = (pw_destroy_t) pw_factory_info_free;
-	} else if (strcmp(type, PW_TYPE_INTERFACE_Client) == 0) {
+	} else if (spa_streq(type, PW_TYPE_INTERFACE_Client)) {
 		events = &client_events;
 		client_version = PW_VERSION_CLIENT;
 		destroy = (pw_destroy_t) pw_client_info_free;
-	} else if (strcmp(type, PW_TYPE_INTERFACE_Link) == 0) {
+	} else if (spa_streq(type, PW_TYPE_INTERFACE_Link)) {
 		events = &link_events;
 		client_version = PW_VERSION_LINK;
 		destroy = (pw_destroy_t) pw_link_info_free;
@@ -613,15 +657,15 @@ static void registry_event_global(void *data, uint32_t id,
 		printf("\tpermissions: "PW_PERMISSION_FORMAT"\n",
 				PW_PERMISSION_ARGS(permissions));
 		printf("\ttype: %s (version %d)\n", type, version);
-		print_properties(props, ' ');
+		print_properties(props, false);
 		return;
 	}
 
-        proxy = pw_registry_bind(d->registry, id, type,
+	proxy = pw_registry_bind(d->registry, id, type,
 				       client_version,
 				       sizeof(struct proxy_data));
-        if (proxy == NULL)
-                goto no_mem;
+	if (proxy == NULL)
+		goto no_mem;
 
 	pd = pw_proxy_get_user_data(proxy);
 	pd->data = d;
@@ -635,19 +679,40 @@ static void registry_event_global(void *data, uint32_t id,
 	pd->pending_seq = 0;
 	pd->print_func = print_func;
 	spa_list_init(&pd->param_list);
-        pw_proxy_add_object_listener(proxy, &pd->object_listener, events, pd);
-        pw_proxy_add_listener(proxy, &pd->proxy_listener, &proxy_events, pd);
-        return;
+	pw_proxy_add_object_listener(proxy, &pd->object_listener, events, pd);
+	pw_proxy_add_listener(proxy, &pd->proxy_listener, &proxy_events, pd);
+	spa_list_append(&d->global_list, &pd->global_link);
 
-      no_mem:
-        printf("failed to create proxy");
-        return;
+	return;
+
+no_mem:
+	fprintf(stderr, "failed to create proxy");
+	return;
+}
+
+static struct proxy_data *find_proxy(struct data *d, uint32_t id)
+{
+	struct proxy_data *pd;
+	spa_list_for_each(pd, &d->global_list, global_link) {
+		if (pd->id == id)
+			return pd;
+	}
+	return NULL;
 }
 
 static void registry_event_global_remove(void *object, uint32_t id)
 {
+	struct data *d = object;
+	struct proxy_data *pd;
+
 	printf("removed:\n");
 	printf("\tid: %u\n", id);
+
+	pd = find_proxy(d, id);
+	if (pd == NULL)
+		return;
+	if (pd->proxy)
+		pw_proxy_destroy(pd->proxy);
 }
 
 static const struct pw_registry_events registry_events = {
@@ -680,12 +745,14 @@ static void do_quit(void *data, int signal_number)
 	pw_main_loop_quit(d->loop);
 }
 
-static void show_help(const char *name)
+static void show_help(const char *name, bool error)
 {
-        fprintf(stdout, "%s [options]\n"
+        fprintf(error ? stderr : stdout, "%s [options]\n"
 		"  -h, --help                            Show this help\n"
 		"      --version                         Show version\n"
-		"  -r, --remote                          Remote daemon name\n",
+		"  -r, --remote                          Remote daemon name\n"
+		"  -N, --no-colors                       disable color output\n"
+		"  -C, --color[=WHEN]                    whether to enable color support. WHEN is `never`, `always`, or `auto`\n",
 		name);
 }
 
@@ -698,19 +765,28 @@ int main(int argc, char *argv[])
 		{ "help",	no_argument,		NULL, 'h' },
 		{ "version",	no_argument,		NULL, 'V' },
 		{ "remote",	required_argument,	NULL, 'r' },
+		{ "no-colors",	no_argument,		NULL, 'N' },
+		{ "color",	optional_argument,	NULL, 'C' },
 		{ NULL,	0, NULL, 0}
 	};
 	int c;
+	bool colors = false;
 
+	setlocale(LC_ALL, "");
 	pw_init(&argc, &argv);
 
+	setlinebuf(stdout);
+
+	if (isatty(STDERR_FILENO) && getenv("NO_COLOR") == NULL)
+		colors = true;
+
 	while ((c = getopt_long(argc, argv, "hVr:", long_options, NULL)) != -1) {
 		switch (c) {
 		case 'h':
-			show_help(argv[0]);
+			show_help(argv[0], false);
 			return 0;
 		case 'V':
-			fprintf(stdout, "%s\n"
+			printf("%s\n"
 				"Compiled with libpipewire %s\n"
 				"Linked with libpipewire %s\n",
 				argv[0],
@@ -720,12 +796,34 @@ int main(int argc, char *argv[])
 		case 'r':
 			opt_remote = optarg;
 			break;
+		case 'N' :
+			colors = false;
+			break;
+		case 'C' :
+			if (optarg == NULL || !strcmp(optarg, "auto"))
+				break; /* nothing to do, tty detection was done
+					  before parsing options */
+			else if (!strcmp(optarg, "never"))
+				colors = false;
+			else if (!strcmp(optarg, "always"))
+				colors = true;
+			else {
+				fprintf(stderr, "Invalid color: %s\n", optarg);
+				show_help(argv[0], true);
+				return -1;
+			}
+			break;
 		default:
-			show_help(argv[0]);
+			show_help(argv[0], true);
 			return -1;
 		}
 	}
 
+	if (colors) {
+		pprefix[1].prefix = SPA_ANSI_RED "*";
+		pprefix[1].suffix = SPA_ANSI_RESET;
+	}
+
 	data.loop = pw_main_loop_new(NULL);
 	if (data.loop == NULL) {
 		fprintf(stderr, "can't create main loop: %m\n");
@@ -743,6 +841,7 @@ int main(int argc, char *argv[])
 	}
 
 	spa_list_init(&data.pending_list);
+	spa_list_init(&data.global_list);
 
 	data.core = pw_context_connect(data.context,
 			pw_properties_new(
@@ -765,7 +864,9 @@ int main(int argc, char *argv[])
 
 	pw_main_loop_run(data.loop);
 
+	spa_hook_remove(&data.registry_listener);
 	pw_proxy_destroy((struct pw_proxy*)data.registry);
+	spa_hook_remove(&data.core_listener);
 	pw_context_destroy(data.context);
 	pw_main_loop_destroy(data.loop);
 	pw_deinit();
